from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, Union

from omegaconf import MISSING, MissingMandatoryValue


@dataclass
class ColorJitter:
    brightness: Optional[float] = 0.25
    contrast: Optional[float] = 0.25
    saturation: Optional[float] = 0.25
    hue: Optional[float] = 0.1
    colorjitter_p: Optional[float] = 0.5


@dataclass
class AugmentationConfig:
    img_size: int = 256
    max_scale: Optional[int] = 1024
    min_scale: Optional[int] = None
    crop_size_h: Optional[int] = None
    crop_size_w: Optional[int] = None
    resize_ratio0: Optional[float] = None
    resize_ratiof: Optional[float] = None
    resize_add: Optional[float] = 1
    fliplr: Optional[float] = 0.5
    color_jitter: Optional[ColorJitter] = field(default_factory=lambda: ColorJitter())
    
    

@dataclass
class ClassificationAugmentationConfig(AugmentationConfig):
    resize_ratio0 = None
    resize_ratiof = None
    resize_add = None
    color_jitter = None


@dataclass
class SegmentationAugmentationConfig(AugmentationConfig):
    img_size = 512
    resize_ratio0 = 1.0
    resize_ratiof = 1.5
    
    def __post_init__(self):
        # variable interpolation
        if self.min_scale is None:
            self.min_scale = self.img_size
        if self.crop_size_h is None:
            self.crop_size_h = self.img_size
        if self.crop_size_w is None:
            self.crop_size_w = self.img_size
    

@dataclass
class DetectionAugmentationConfig(AugmentationConfig):
    img_size = 512
    max_scale = 2048
    min_scale = 768
    resize_ratio0: 0.5
    resize_ratiof: 2.0
    resize_add: 1
    
    def __post_init__(self):
        # variable interpolation
        if self.crop_size_h is None:
            self.crop_size_h = self.img_size
        if self.crop_size_w is None:
            self.crop_size_w = self.img_size


import logging
import os
from pathlib import Path
from typing import Callable, Dict, List, Type

import torch
import torch.nn as nn
from omegaconf import OmegaConf

from .base import ClassificationModel, DetectionModel, SegmentationModel, TaskModel
from .registry import MODEL_FULL_DICT, SUPPORTING_TASK_LIST
from .utils import load_from_checkpoint

logger = logging.getLogger("netspresso_trainer")


def load_full_model(conf_model, model_name, num_classes, model_checkpoint):
    model_fn: Callable[..., nn.Module] = MODEL_FULL_DICT[model_name]
    conf_model_full = OmegaConf.to_object(conf_model.architecture.full)
    model: nn.Module = model_fn(num_classes=num_classes, conf_model_full=conf_model_full)
    model = load_from_checkpoint(model, model_checkpoint)

    return model


def load_backbone_and_head_model(conf_model, task, backbone_name, head_name, num_classes, model_checkpoint, img_size, freeze_backbone):
    TASK_MODEL_DICT: Dict[str, Type[TaskModel]] = {
        'classification': ClassificationModel,
        'segmentation': SegmentationModel,
        'detection': DetectionModel
    }

    if task not in TASK_MODEL_DICT:
        raise ValueError(f"No such task(s) named: {task}. This should be included in SUPPORTING_TASK_LIST ({SUPPORTING_TASK_LIST})")

    return TASK_MODEL_DICT[task](conf_model, task, backbone_name, head_name, num_classes, model_checkpoint, img_size, freeze_backbone)


def build_model(conf_model, task, num_classes, model_checkpoint, img_size) -> nn.Module:

    if conf_model.single_task_model:
        model_name = str(conf_model.architecture.full.name).lower()
        return load_full_model(conf_model, model_name, num_classes, model_checkpoint)

    backbone_name = str(conf_model.architecture.backbone.name).lower()
    head_name = str(conf_model.architecture.head.name).lower()
    freeze_backbone = conf_model.freeze_backbone
    return load_backbone_and_head_model(conf_model, task, backbone_name, head_name, num_classes, model_checkpoint, img_size, freeze_backbone)



import logging
import math
import warnings

import torch
from torch.optim.lr_scheduler import _LRScheduler

logger = logging.getLogger("netspresso_trainer")

class CosineAnnealingLRWithCustomWarmUp(_LRScheduler):
    """Modified from CosineAnnealingLR in PyTorch

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        warmup_iters (int): The number of steps that the scheduler finishes to warmup the learning rate. Default: 5.
        total_iters (int): Maximum number of iterations. Originally named as `T_max`. Default: 5.
        warmup_bias_lr (int): Starting learning rate for warmup period. Default: 0.
        min_lr (float): Minimum learning rate. Originally named as `eta_min`. Default: 0.
        last_epoch (int): The index of last epoch. Default: -1.
        verbose (bool): If ``True``, prints a message to stdout for each update. Default: ``False``.
    """

    def __init__(self, optimizer, warmup_iters=5, total_iters=5, warmup_bias_lr=0, min_lr=0, last_epoch=-1, verbose=False, **kwargs):
        self.T_max = total_iters
        self.eta_min = min_lr
        self.warmup_bias_lr = warmup_bias_lr
        self.warmup_iters = warmup_iters
        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning, stacklevel=2)
        
        if self.last_epoch > self.T_max:
            return [group['lr'] for group in self.optimizer.param_groups]
        
        if self.last_epoch >= 0 and self.last_epoch < self.warmup_iters:
            return [self.warmup_bias_lr + (float(self.last_epoch + 1) / float(max(1, self.warmup_iters))) * (base_lr - self.warmup_bias_lr)
                    for base_lr in self.base_lrs]

        if self._step_count == 1 and self.last_epoch > 0:
            return [self.eta_min + (base_lr - self.eta_min) *
                    (1 + math.cos((self.last_epoch) * math.pi / self.T_max)) / 2
                    for base_lr, group in
                    zip(self.base_lrs, self.optimizer.param_groups)]
        elif (self.last_epoch - 1 - self.T_max) % (2 * self.T_max) == 0:
            return [group['lr'] + (base_lr - self.eta_min) *
                    (1 - math.cos(math.pi / self.T_max)) / 2
                    for base_lr, group in
                    zip(self.base_lrs, self.optimizer.param_groups)]
        return [(1 + math.cos(math.pi * self.last_epoch / self.T_max)) /
                (1 + math.cos(math.pi * (self.last_epoch - 1) / self.T_max)) *
                (group['lr'] - self.eta_min) + self.eta_min
                for group in self.optimizer.param_groups]

    def _get_closed_form_lr(self):
        return [
            (
                min(
                    self.warmup_bias_lr + (float(self.last_epoch + 1) / float(max(1, self.warmup_iters))) * (base_lr - self.warmup_bias_lr),
                    self.eta_min + (base_lr - self.eta_min) * (1 + math.cos(math.pi * self.last_epoch / self.T_max)) / 2
                )
            )
            for base_lr in self.base_lrs
        ]


import logging
from pathlib import Path
from typing import Any, List, Optional, TypedDict, Union

import omegaconf
import torch
import torch.nn as nn
from torch import Tensor
from torch.fx.proxy import Proxy

logger = logging.getLogger("netspresso_trainer")

FXTensorType = Union[Tensor, Proxy]
FXTensorListType = Union[List[Tensor], List[Proxy]]


class BackboneOutput(TypedDict):
    intermediate_features: Optional[FXTensorListType]
    last_feature: Optional[FXTensorType]


class ModelOutput(TypedDict):
    pred: FXTensorType


class DetectionModelOutput(ModelOutput):
    boxes: Any
    proposals: Any
    anchors: Any
    objectness: Any
    pred_bbox_detlas: Any
    class_logits: Any
    box_regression: Any
    labels: Any
    regression_targets: Any
    post_boxes: Any
    post_scores: Any
    post_labels: Any


class PIDNetModelOutput(ModelOutput):
    extra_p: Optional[FXTensorType]
    extra_d: Optional[FXTensorType]


def load_from_checkpoint(model: nn.Module, model_checkpoint: Optional[Union[str, Path]]) -> nn.Module:
    if model_checkpoint is not None:
        model_state_dict = torch.load(model_checkpoint, map_location='cpu')
        missing_keys, unexpected_keys = model.load_state_dict(model_state_dict, strict=False)

        if len(missing_keys) != 0:
            logger.warning(f"Missing key(s) in state_dict: {missing_keys}")
        if len(unexpected_keys) != 0:
            logger.warning(f"Unexpected key(s) in state_dict: {unexpected_keys}")

    return model


def is_single_task_model(conf_model: omegaconf.DictConfig):
    conf_model_architecture_full = conf_model.architecture.full
    if conf_model_architecture_full is None:
        return False
    if conf_model_architecture_full.name is None:
        return False
    return True



import torch
import torch.nn as nn
import torch.nn.functional as F

from ....op.custom import ConvLayer, CSPLayer


class FPN(nn.Module):

    def __init__(self,
                 in_channels,
                 out_channels,
                 num_outs,
                 start_level=0,
                 end_level=-1,
                 add_extra_convs=False,
                 relu_before_extra_convs=False,
                 no_norm_on_lateral=False,
                 conv_cfg=None,
                 norm_cfg=None,
                 act_cfg=None,
                 upsample_cfg=None,
                 init_cfg=None):
        if init_cfg is None:
            init_cfg = {'type': 'Xavier', 'layer': 'Conv2d', 'distribution': 'uniform'}
        if upsample_cfg is None:
            upsample_cfg = {'mode': 'nearest'}
        super(FPN, self).__init__()
        assert isinstance(in_channels, list)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_ins = len(in_channels)
        self.num_outs = num_outs
        self.relu_before_extra_convs = relu_before_extra_convs
        self.no_norm_on_lateral = no_norm_on_lateral
        self.fp16_enabled = False
        self.upsample_cfg = upsample_cfg.copy()

        if end_level == -1 or end_level == self.num_ins - 1:
            self.backbone_end_level = self.num_ins
            assert num_outs >= self.num_ins - start_level
        else:
            # if end_level is not the last level, no extra level is allowed
            self.backbone_end_level = end_level + 1
            assert end_level < self.num_ins
            assert num_outs == end_level - start_level + 1
        self.start_level = start_level
        self.end_level = end_level
        self.add_extra_convs = add_extra_convs
        assert isinstance(add_extra_convs, (str, bool))
        if isinstance(add_extra_convs, str):
            # Extra_convs_source choices: 'on_input', 'on_lateral', 'on_output'
            assert add_extra_convs in ('on_input', 'on_lateral', 'on_output')
        elif add_extra_convs:  # True
            self.add_extra_convs = 'on_input'

        self.lateral_convs = nn.ModuleList()
        self.fpn_convs = nn.ModuleList()

        for i in range(self.start_level, self.backbone_end_level):
            l_conv = nn.Conv2d(in_channels[i], out_channels, kernel_size=1, stride=1, padding=0)
            # ConvModule(
            #     in_channels[i],
            #     out_channels,
            #     1,
            #     conv_cfg=conv_cfg,
            #     norm_cfg=norm_cfg if not self.no_norm_on_lateral else None,
            #     act_cfg=act_cfg,
            #     inplace=False)
            fpn_conv = nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1)
            # ConvModule(
            #     out_channels,
            #     out_channels,
            #     3,
            #     padding=1,
            #     conv_cfg=conv_cfg,
            #     norm_cfg=norm_cfg,
            #     act_cfg=act_cfg,
            #     inplace=False)

            self.lateral_convs.append(l_conv)
            self.fpn_convs.append(fpn_conv)

        # add extra conv layers (e.g., RetinaNet)
        extra_levels = num_outs - self.backbone_end_level + self.start_level
        if self.add_extra_convs and extra_levels >= 1:
            for i in range(extra_levels):
                if i == 0 and self.add_extra_convs == 'on_input':
                    in_channels = self.in_channels[self.backbone_end_level - 1]
                else:
                    in_channels = out_channels
                extra_fpn_conv = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=2, padding=1)
                self.fpn_convs.append(extra_fpn_conv)

    def forward(self, inputs):
        """Forward function."""
        assert len(inputs) == len(self.in_channels)

        # build laterals
        laterals = [
            lateral_conv(inputs[i + self.start_level])
            for i, lateral_conv in enumerate(self.lateral_convs)
        ]

        # build top-down path
        used_backbone_levels = len(laterals)
        for i in range(used_backbone_levels - 1, 0, -1):
            # In some cases, fixing `scale factor` (e.g. 2) is preferred, but
            #  it cannot co-exist with `size` in `F.interpolate`.
            if 'scale_factor' in self.upsample_cfg:
                # fix runtime error of "+=" inplace operation in PyTorch 1.10
                laterals[i - 1] = laterals[i - 1] + F.interpolate(
                    laterals[i], **self.upsample_cfg)
            else:
                prev_shape = laterals[i - 1].shape[2:]
                laterals[i - 1] = laterals[i - 1] + F.interpolate(
                    laterals[i], size=prev_shape, **self.upsample_cfg)

        # build outputs
        # part 1: from original levels
        outs = [
            self.fpn_convs[i](laterals[i]) for i in range(used_backbone_levels)
        ]
        # part 2: add extra levels
        if self.num_outs > len(outs):
            # use max pool to get more levels on top of outputs
            # (e.g., Faster R-CNN, Mask R-CNN)
            if not self.add_extra_convs:
                for i in range(self.num_outs - used_backbone_levels):
                    outs.append(F.max_pool2d(outs[-1], 1, stride=2))
            # add conv layers on top of original feature maps (RetinaNet)
            else:
                if self.add_extra_convs == 'on_input':
                    extra_source = inputs[self.backbone_end_level - 1]
                elif self.add_extra_convs == 'on_lateral':
                    extra_source = laterals[-1]
                elif self.add_extra_convs == 'on_output':
                    extra_source = outs[-1]
                else:
                    raise NotImplementedError
                outs.append(self.fpn_convs[used_backbone_levels](extra_source))
                for i in range(used_backbone_levels + 1, self.num_outs):
                    if self.relu_before_extra_convs:
                        outs.append(self.fpn_convs[i](F.relu(outs[-1])))
                    else:
                        outs.append(self.fpn_convs[i](outs[-1]))
        return outs


class PAFPN(nn.Module):
    """
    YOLOv3 model. Darknet 53 is the default backbone of this model.
    """

    def __init__(
        self,
        in_channels,
        act_type="silu",
    ):
        super().__init__()
        
        self.in_channels = in_channels
        Conv = ConvLayer

        # TODO: Get from config
        depth = 0.33

        self.upsample = nn.Upsample(scale_factor=2, mode="nearest")
        self.lateral_conv0 = ConvLayer(
            in_channels=int(in_channels[2]), 
            out_channels=int(in_channels[1]), 
            kernel_size=1, 
            stride=1,
            act_type=act_type
        )
        self.C3_p4 = CSPLayer(
            in_channels=int(2 * in_channels[1]),
            out_channels=int(in_channels[1]),
            n=round(3 * depth),
            shortcut=False,
            act_type=act_type,
        )  # cat

        self.reduce_conv1 = ConvLayer(
            in_channels=int(in_channels[1]), 
            out_channels=int(in_channels[0]), 
            kernel_size=1, 
            stride=1, 
            act_type=act_type
        )
        self.C3_p3 = CSPLayer(
            in_channels=int(2 * in_channels[0]),
            out_channels=int(in_channels[0]),
            n=round(3 * depth),
            shortcut=False,
            act_type=act_type,
        )

        # bottom-up conv
        self.bu_conv2 = Conv(
            in_channels=int(in_channels[0]), 
            out_channels=int(in_channels[0]), 
            kernel_size=3, 
            stride=2, 
            act_type=act_type
        )
        self.C3_n3 = CSPLayer(
            in_channels=int(2 * in_channels[0]),
            out_channels=int(in_channels[1]),
            n=round(3 * depth),
            shortcut=False,
            act_type=act_type,
        )

        # bottom-up conv
        self.bu_conv1 = Conv(
            in_channels=int(in_channels[1]), 
            out_channels=int(in_channels[1]), 
            kernel_size=3, 
            stride=2, 
            act_type=act_type
        )
        self.C3_n4 = CSPLayer(
            in_channels=int(2 * in_channels[1]),
            out_channels=int(in_channels[2]),
            n=round(3 * depth),
            shortcut=False,
            act_type=act_type,
        )

    def forward(self, inputs):
        """
        Args:
            inputs: input images.

        Returns:
            Tuple[Tensor]: FPN feature.
        """

        [x2, x1, x0] = inputs

        fpn_out0 = self.lateral_conv0(x0)  # 1024->512/32
        f_out0 = self.upsample(fpn_out0)  # 512/16
        f_out0 = torch.cat([f_out0, x1], 1)  # 512->1024/16
        f_out0 = self.C3_p4(f_out0)  # 1024->512/16

        fpn_out1 = self.reduce_conv1(f_out0)  # 512->256/16
        f_out1 = self.upsample(fpn_out1)  # 256/8
        f_out1 = torch.cat([f_out1, x2], 1)  # 256->512/8
        pan_out2 = self.C3_p3(f_out1)  # 512->256/8

        p_out1 = self.bu_conv2(pan_out2)  # 256->256/16
        p_out1 = torch.cat([p_out1, fpn_out1], 1)  # 256->512/16
        pan_out1 = self.C3_n3(p_out1)  # 512->512/16

        p_out0 = self.bu_conv1(pan_out1)  # 512->512/32
        p_out0 = torch.cat([p_out0, fpn_out0], 1)  # 512->1024/32
        pan_out0 = self.C3_n4(p_out0)  # 1024->1024/32

        outputs = (pan_out2, pan_out1, pan_out0)
        return outputs



from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import PIL.Image as Image


class ImageSaver:
    def __init__(self, model, result_dir) -> None:
        super(ImageSaver, self).__init__()
        self.model = model
        self.save_dir: Path = Path(result_dir) / "result_image"
        self.save_dir.mkdir(exist_ok=True)
        self._epoch = None
    
    def init_epoch(self):
        self._epoch = 0
        
    @property
    def epoch(self):
        return self._epoch
    
    @epoch.setter
    def epoch(self, value: int) -> None:
        self._epoch = int(value)
        
    def save_ndarray_as_image(self, image_array: np.ndarray, filename: Union[str, Path], dataformats: Literal['HWC', 'CHW'] = 'HWC'):
        assert image_array.ndim == 3
        if dataformats != 'HWC' and dataformats == 'CHW':
            image_array = image_array.transpose((1, 2, 0))
        
        # HWC
        assert image_array.shape[-1] in [1, 3]
        Image.fromarray(image_array.astype(np.uint8)).save(filename)
        return True
        
    def save_result(self, image_dict: Dict, prefix='train'):
        prefix_dir: Path = self.save_dir / prefix
        prefix_dir.mkdir(exist_ok=True)
        
        for k, v in image_dict.items():
            assert isinstance(v, np.ndarray)
            assert v.ndim in [3, 4], \
                f"Array for saving as image should have dim of 3 or 4! Current: {v.ndim}"
            if v.ndim == 3:
                self.save_ndarray_as_image(v, f"{prefix_dir}/{self._epoch:04d}_{k}.png", dataformats='HWC')
            for idx, image in enumerate(v):
                filename = f"{prefix_dir}/{self._epoch:04d}_{idx:03d}_{k}.png"
                self.save_ndarray_as_image(image, filename, dataformats='HWC') # TODO: get dataformats option from outside

    def __call__(self, train_images=None, valid_images=None):
        if train_images is not None:
            self.save_result(train_images, prefix='train')
        if valid_images is not None:
            self.save_result(valid_images, prefix='valid')
        




from pathlib import Path
from typing import Callable, Dict, List, Type

import torch.nn as nn

from .backbones import cspdarknet, efficientformer, mobilenetv3_small, mobilevit, resnet50, segformer, vit
from .full import pidnet
from .heads.classification import fc
from .heads.detection import faster_rcnn, yolo_head
from .heads.segmentation import all_mlp_decoder

MODEL_BACKBONE_DICT: Dict[str, Callable[..., nn.Module]] = {
    'resnet50': resnet50,
    'mobilenetv3_small': mobilenetv3_small,
    'segformer': segformer,
    'mobilevit': mobilevit,
    'vit': vit,
    'efficientformer': efficientformer,
    'cspdarknet': cspdarknet
}

MODEL_HEAD_DICT: Dict[str, Callable[..., nn.Module]] = {
    'classification': {
        'fc': fc,
    },
    'segmentation': {
        'all_mlp_decoder': all_mlp_decoder,
    },
    'detection': {
        'faster_rcnn': faster_rcnn,
        'yolo_head': yolo_head
    },
}

MODEL_FULL_DICT = {
    'pidnet': pidnet
}

SUPPORTING_MODEL_LIST = list(MODEL_BACKBONE_DICT.keys()) + list(MODEL_FULL_DICT.keys())
SUPPORTING_TASK_LIST: List[str] = ['classification', 'segmentation', 'detection']



from typing import Optional

import cv2
import numpy as np
import PIL.Image as Image

from ..augmentation import custom as TC
from ..utils.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD


def train_transforms_efficientformer(conf_augmentation):
    
    crop_size_h = conf_augmentation.crop_size_h
    crop_size_w = conf_augmentation.crop_size_w
    
    train_transforms_composed = TC.Compose([
        TC.Resize(size=(crop_size_h, crop_size_w)),
        TC.ToTensor(),
        TC.Normalize(mean=IMAGENET_DEFAULT_MEAN, std=IMAGENET_DEFAULT_STD)
    ])

    return train_transforms_composed

def val_transforms_efficientformer(conf_augmentation):
    
    crop_size_h = conf_augmentation.crop_size_h
    crop_size_w = conf_augmentation.crop_size_w
    
    val_transforms_composed = TC.Compose([
        TC.Resize(size=(crop_size_h, crop_size_w)),
        TC.ToTensor(),
        TC.Normalize(mean=IMAGENET_DEFAULT_MEAN, std=IMAGENET_DEFAULT_STD)
    ])

    return val_transforms_composed

def create_transform_detection(model_name: str, is_training=False):
    if is_training:
        return train_transforms_efficientformer
    return val_transforms_efficientformer



from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F


class CrossEntropyLoss(nn.Module):
    def __init__(self, ignore_index, **kwargs) -> None:
        super(CrossEntropyLoss, self).__init__()
        self.loss_fn = nn.CrossEntropyLoss(ignore_index=ignore_index, **kwargs)

    def forward(self, out: Dict, target: torch.Tensor) -> torch.Tensor:
        pred = out['pred']
        loss = self.loss_fn(pred, target)
        return loss


from typing import Dict, List, Optional, Tuple, Union

from torch import nn, Tensor
from torchvision.utils import _log_api_usage_once
from torchvision.ops.poolers import LevelMapper, _filter_input, _setup_scales, _multiscale_roi_align


class MultiScaleRoIAlign(nn.Module):
    """
    Original class is torchvision.ops.poolers.MultiScaleRoIAlign
    """

    __annotations__ = {"scales": Optional[List[float]], "map_levels": Optional[LevelMapper]}

    def __init__(
        self,
        featmap_names: List[str],
        output_size: Union[int, Tuple[int], List[int]],
        sampling_ratio: int,
        *,
        canonical_scale: int = 224,
        canonical_level: int = 4,
    ):
        super().__init__()
        _log_api_usage_once(self)
        if isinstance(output_size, int):
            output_size = (output_size, output_size)
        self.featmap_names = featmap_names
        self.sampling_ratio = sampling_ratio
        self.output_size = tuple(output_size)
        self.scales = None
        self.map_levels = None
        self.canonical_scale = canonical_scale
        self.canonical_level = canonical_level

    def forward(
        self,
        x: Dict[str, Tensor],
        boxes: List[Tensor],
        image_shapes: Tuple[int, int],
    ) -> Tensor:
        """
        Args:
            x (OrderedDict[Tensor]): feature maps for each level. They are assumed to have
                all the same number of channels, but they can have different sizes.
            boxes (List[Tensor[N, 4]]): boxes to be used to perform the pooling operation, in
                (x1, y1, x2, y2) format and in the image reference size, not the feature map
                reference. The coordinate must satisfy ``0 <= x1 < x2`` and ``0 <= y1 < y2``.
            image_shapes (List[Tuple[height, width]]): the sizes of each image before they
                have been fed to a CNN to obtain feature maps. This allows us to infer the
                scale factor for each one of the levels to be pooled.
        Returns:
            result (Tensor)
        """
        x_filtered = _filter_input(x, self.featmap_names)
        if self.scales is None or self.map_levels is None:
            image_shapes = [image_shapes for _ in range(len(x_filtered[0]))]
            self.scales, self.map_levels = _setup_scales(
                x_filtered, image_shapes, self.canonical_scale, self.canonical_level
            )

        return _multiscale_roi_align(
            x_filtered,
            boxes,
            self.output_size,
            self.sampling_ratio,
            self.scales,
            self.map_levels,
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(featmap_names={self.featmap_names}, "
            f"output_size={self.output_size}, sampling_ratio={self.sampling_ratio})"
        )



from typing import Dict, List, Type

from .base import BasePipeline
from .classification import ClassificationPipeline
from .detection import OneStageDetectionPipeline, TwoStageDetectionPipeline
from .segmentation import SegmentationPipeline

# TODO: Temporary defined. It should be integrated with `..models.registry.SUPPORTING_TASK_LIST`
SUPPORTING_TASK_LIST: List[str] = ['classification', 'segmentation', 'detection']

TASK_PIPELINE: Dict[str, Type[BasePipeline]]= {
    'classification': ClassificationPipeline,
    'segmentation': SegmentationPipeline,
    'detection-two-stage': TwoStageDetectionPipeline,
    'detection-one-stage': OneStageDetectionPipeline,
}


from typing import Optional

import cv2
import numpy as np
import PIL.Image as Image

from ..augmentation import custom as TC
from ..utils.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

EDGE_SIZE = 4
Y_K_SIZE = 6
X_K_SIZE = 6


def reduce_label(label: np.ndarray) -> Image.Image:
    label[label == 0] = 255
    label = label - 1
    label[label == 254] = 255
    return Image.fromarray(label)

def generate_edge(label: np.ndarray) -> Image.Image:
    edge = cv2.Canny(label, 0.1, 0.2)
    kernel = np.ones((EDGE_SIZE, EDGE_SIZE), np.uint8)
    # edge_pad == True
    edge = edge[Y_K_SIZE:-Y_K_SIZE, X_K_SIZE:-X_K_SIZE]
    edge = np.pad(edge, ((Y_K_SIZE, Y_K_SIZE), (X_K_SIZE, X_K_SIZE)), mode='constant')
    edge = (cv2.dilate(edge, kernel, iterations=1) > 50) * 1.0
    return Image.fromarray((edge.copy() * 255).astype(np.uint8))


def train_transforms_segmentation(conf_augmentation):

    crop_size_h = conf_augmentation.crop_size_h
    crop_size_w = conf_augmentation.crop_size_w

    scale_ratio = (conf_augmentation.resize_ratio0, conf_augmentation.resize_ratiof)
    
    train_transforms_composed = TC.Compose([
        TC.RandomResizedCrop((crop_size_h, crop_size_w), scale=scale_ratio, ratio=(1.0, 1.0)),
        TC.RandomHorizontalFlip(p=conf_augmentation.fliplr),
        TC.ColorJitter(brightness=conf_augmentation.color_jitter.brightness,
                       contrast=conf_augmentation.color_jitter.contrast,
                       saturation=conf_augmentation.color_jitter.saturation,
                       hue=conf_augmentation.color_jitter.hue,
                       p=conf_augmentation.color_jitter.colorjitter_p),
        TC.ToTensor(),
        TC.Normalize(mean=IMAGENET_DEFAULT_MEAN, std=IMAGENET_DEFAULT_STD)
    ])

    return train_transforms_composed

def val_transforms_segmentation(conf_augmentation):

    crop_size_h = conf_augmentation.crop_size_h
    crop_size_w = conf_augmentation.crop_size_w

    val_transforms_composed = TC.Compose([
        TC.Resize((crop_size_h, crop_size_w)),
        TC.ToTensor(),
        TC.Normalize(mean=IMAGENET_DEFAULT_MEAN, std=IMAGENET_DEFAULT_STD)
    ])

    return val_transforms_composed


def infer_transforms_segmentation(conf_augmentation):
    return


def train_transforms_pidnet(conf_augmentation):

    crop_size_h = conf_augmentation.crop_size_h
    crop_size_w = conf_augmentation.crop_size_w

    scale_ratio = (conf_augmentation.resize_ratio0, conf_augmentation.resize_ratiof)

    train_transforms_composed = TC.Compose(
        [
            TC.RandomResizedCrop((crop_size_h, crop_size_w), scale=scale_ratio, ratio=(1.0, 1.0)),
            TC.RandomHorizontalFlip(p=conf_augmentation.fliplr),
            TC.ToTensor(),
            TC.Normalize(mean=IMAGENET_DEFAULT_MEAN, std=IMAGENET_DEFAULT_STD)
        ],
        additional_targets={'edge': 'mask'}
    )
    
    return train_transforms_composed


def val_transforms_pidnet(conf_augmentation):

    crop_size_h = conf_augmentation.crop_size_h
    crop_size_w = conf_augmentation.crop_size_w

    val_transforms_composed = TC.Compose(
        [
            TC.Resize((crop_size_h, crop_size_w)),
            TC.ToTensor(),
            TC.Normalize(mean=IMAGENET_DEFAULT_MEAN, std=IMAGENET_DEFAULT_STD)
        ],
        additional_targets={'edge': 'mask'}
    )

    return val_transforms_composed


def infer_transforms_pidnet(conf_augmentation):
    return


def create_transform_segmentation(model_name: str, is_training=False):

    if model_name == 'pidnet':
        if is_training:
            return train_transforms_pidnet
        return val_transforms_pidnet
    if is_training:
        return train_transforms_segmentation
    return val_transforms_segmentation



import logging
import os
from abc import abstractmethod
from typing import Callable, Optional, Tuple, Union

import torch
import torch.nn as nn
from omegaconf import OmegaConf

from .registry import MODEL_BACKBONE_DICT, MODEL_HEAD_DICT
from .utils import BackboneOutput, DetectionModelOutput, ModelOutput, load_from_checkpoint

logger = logging.getLogger("netspresso_trainer")


class TaskModel(nn.Module):
    def __init__(self, conf_model, task, backbone_name, head_name, num_classes, model_checkpoint,
                 img_size: Optional[Union[int, Tuple]] = None, freeze_backbone: bool = False) -> None:
        super(TaskModel, self).__init__()
        self.task = task
        self.backbone_name = backbone_name
        self.head_name = head_name

        backbone_fn: Callable[..., nn.Module] = MODEL_BACKBONE_DICT[backbone_name]
        conf_model_backbone = OmegaConf.to_object(conf_model.architecture.backbone)
        self.backbone: nn.Module = backbone_fn(task=self.task, conf_model_backbone=conf_model_backbone)

        self.backbone = load_from_checkpoint(self.backbone, model_checkpoint)

        head_module = MODEL_HEAD_DICT[self.task][head_name]
        if task == 'classification':
            self.head = head_module(num_classes=num_classes, feature_dim=self.backbone.feature_dim)
        elif task in ['segmentation', 'detection']:
            img_size = img_size if isinstance(img_size, (int, None)) else tuple(img_size)
            self.head = head_module(num_classes=num_classes,
                                    intermediate_features_dim=self.backbone.intermediate_features_dim,
                                    label_size=img_size)

        if freeze_backbone:
            self._freeze_backbone()
            logger.info(f"Freeze! {self.backbone_name} is now freezed. Now only tuning with {self.head_name}.")

    def _freeze_backbone(self):
        for m in self.backbone.parameters():
            m.requires_grad = False

    @property
    def device(self):
        return next(self.parameters()).device

    def _get_name(self):
        return f"{self.__class__.__name__}[task={self.task}, backbone={self.backbone_name}, head={self.head_name}]"

    @abstractmethod
    def forward(self, x, label_size=None, targets=None):
        raise NotImplementedError


class ClassificationModel(TaskModel):
    def __init__(self, conf_model, task, backbone_name, head_name, num_classes, model_checkpoint,
                 label_size=None, freeze_backbone=False) -> None:
        super().__init__(conf_model, task, backbone_name, head_name, num_classes, model_checkpoint, label_size, freeze_backbone)

    def forward(self, x, label_size=None, targets=None):
        features: BackboneOutput = self.backbone(x)
        out: ModelOutput = self.head(features['last_feature'])
        return out


class SegmentationModel(TaskModel):
    def __init__(self, conf_model, task, backbone_name, head_name, num_classes, model_checkpoint,
                 label_size, freeze_backbone=False) -> None:
        super().__init__(conf_model, task, backbone_name, head_name, num_classes, model_checkpoint, label_size, freeze_backbone)

    def forward(self, x, label_size=None, targets=None):
        features: BackboneOutput = self.backbone(x)
        out: ModelOutput = self.head(features['intermediate_features'])
        return out


class DetectionModel(TaskModel):
    def __init__(self, conf_model, task, backbone_name, head_name, num_classes, model_checkpoint,
                 label_size, freeze_backbone=False) -> None:
        super().__init__(conf_model, task, backbone_name, head_name, num_classes, model_checkpoint, label_size, freeze_backbone)

    def forward(self, x, label_size=None, targets=None):
        features: BackboneOutput = self.backbone(x)
        out: DetectionModelOutput = self.head(features['intermediate_features'])
        return out




from omegaconf import OmegaConf

from .registry import SCHEDULER_DICT


def build_scheduler(optimizer, conf_training):
    scheduler_name = conf_training.sched
    num_epochs = conf_training.epochs
    conf_sched = OmegaConf.create({
        'min_lr': conf_training.min_lr,
        'power': conf_training.sched_power,
        'warmup_bias_lr': conf_training.warmup_bias_lr,
        'warmup_iters': conf_training.warmup_epochs,
        'total_iters': num_epochs,
        'iters_per_phase': conf_training.iters_per_phase,  # TODO: config for StepLR
    })
    
    assert scheduler_name in SCHEDULER_DICT, f"{scheduler_name} not in scheduler dict!"
    lr_scheduler = SCHEDULER_DICT[scheduler_name](optimizer, **conf_sched)
    
    return lr_scheduler, num_epochs



"""
Implements the Generalized R-CNN framework
"""
from typing import Tuple

from torch import nn

from .....utils import DetectionModelOutput, FXTensorListType


class GeneralizedRCNN(nn.Module):
    """
    Main class for Generalized R-CNN.

    Args:
        backbone (nn.Module):
        rpn (nn.Module):
        roi_heads (nn.Module): takes the features + the proposals from the RPN and computes
            detections / masks from it.
    """

    def __init__(self, neck:nn.Module, rpn: nn.Module, roi_heads: nn.Module, image_size: Tuple[int, int]) -> None:
        super().__init__()
        # _log_api_usage_once(self)
        self.neck = neck
        self.rpn = rpn
        self.roi_heads = roi_heads

        self.image_size = image_size

    def forward(self, features: FXTensorListType) -> DetectionModelOutput:
        if self.neck:
            features = self.neck(features)

        features = {str(k): v for k, v in enumerate(features)}
        rpn_features = self.rpn(features, self.image_size)
        roi_features = self.roi_heads(features, rpn_features['boxes'], self.image_size)

        out_features = DetectionModelOutput()
        out_features.update(rpn_features)
        out_features.update(roi_features)

        return out_features


# from .core import *
from .experimental.darknet import cspdarknet
from .experimental.efficientformer import efficientformer
from .experimental.mobilenetv3 import mobilenetv3_small
from .experimental.mobilevit import mobilevit
from .experimental.resnet import resnet50
from .experimental.segformer import segformer
from .experimental.vit import vit



from typing import List

import numpy as np
import torch

from ...utils.record import AverageMeter
from ..base import BaseMetric

IGNORE_INDEX_NONE_VALUE = -100


class SegmentationMetric(BaseMetric):
    metric_names: List[str] = ['iou', 'pixel_acc']
    primary_metric: str = 'iou'

    def __init__(self, num_classes=None, ignore_index=IGNORE_INDEX_NONE_VALUE):
        super().__init__()
        self.ignore_index = ignore_index if ignore_index is not None else IGNORE_INDEX_NONE_VALUE
        self.K = num_classes

    def intersection_and_union_gpu(self, output, target):

        # 'K' classes, output and target sizes are N or N * L or N * H * W, each value in range 0 to K - 1.
        assert (output.dim() in [1, 2, 3])

        assert output.shape == target.shape
        output = output.view(-1)
        target = target.view(-1)
        output[target == self.ignore_index] = self.ignore_index
        intersection = output[output == target]
        area_intersection = torch.histc(intersection, bins=self.K, min=0, max=self.K-1)
        area_output = torch.histc(output, bins=self.K, min=0, max=self.K-1)
        area_target = torch.histc(target, bins=self.K, min=0, max=self.K-1)
        area_union = area_output + area_target - area_intersection

        intersection, union, target, output = \
            area_intersection.cpu().numpy(), area_union.cpu().numpy(), area_target.cpu().numpy(), area_output.cpu().numpy()

        return {
            'intersection': intersection,
            'union': union,
            'target': target,
            'output': output
        }

    def calibrate(self, pred, target, **kwargs):
        result_dict = {k: AverageMeter(k) for k in self.metric_names}
        B = pred.size(0)

        output_seg = torch.max(pred, dim=1)[1]  # argmax
        metrics = self.intersection_and_union_gpu(output_seg, target)
        result_dict['iou'].update(sum(metrics['intersection']) / (sum(metrics['union']) + 1e-10), n=B)
        result_dict['pixel_acc'].update(sum(metrics['intersection']) / (sum(metrics['target']) + 1e-10), n=B)

        return {k: v.avg for k, v in result_dict.items()}



import logging
import os
from pathlib import Path
from typing import List

import numpy as np
import PIL.Image as Image
import torch
from omegaconf import OmegaConf

from ..base import BaseCustomDataset

ID2LABEL_FILENAME = "id2label.json"
TEMP_COCO_LABEL_FILE = "data/detection/coco.yaml"


def exist_name(candidate, folder_iterable):
    try:
        return list(filter(lambda x: candidate[0] in x, folder_iterable))[0]
    except IndexError:
        return list(filter(lambda x: candidate[1] in x, folder_iterable))[0]


def get_label(label_file: Path):
    target = Path(label_file).read_text()
    
    try:
        target_array = np.array([list(map(float, box.split(' '))) for box in target.split('\n') if box.strip()])
    except ValueError as e:
        print(target)
        raise e
        
    label, boxes = target_array[:, 0], target_array[:, 1:]
    label = label[..., np.newaxis]
    return label, boxes


class DetectionCustomDataset(BaseCustomDataset):

    def __init__(self, conf_data, conf_augmentation, model_name, idx_to_class,
                 split, samples, transform=None, with_label=True, **kwargs):
        super(DetectionCustomDataset, self).__init__(
            conf_data, conf_augmentation, model_name, idx_to_class,
            split, samples, transform, with_label, **kwargs
        )
    
    @staticmethod
    def xywhn2xyxy(original: np.ndarray, w: int, h: int, padw=0, padh=0):
        converted = original.copy()
        # left, top (lt)
        converted[..., 0] = w * (original[..., 0] - original[..., 2] / 2) + padw
        converted[..., 1] = h * (original[..., 1] - original[..., 3] / 2) + padh
        # right, bottom (rb)
        converted[..., 2] = w * (original[..., 0] + original[..., 2] / 2) + padw
        converted[..., 3] = h * (original[..., 1] + original[..., 3] / 2) + padh
        return converted

    def __getitem__(self, index):
        img_path = Path(self.samples[index]['image'])
        ann_path = Path(self.samples[index]['label']) if 'label' in self.samples[index] else None
        img = Image.open(str(img_path)).convert('RGB')

        org_img = img.copy()

        w, h = img.size

        if ann_path is None:
            out = self.transform(self.conf_augmentation)(image=img)
            return {'pixel_values': out['image'], 'name': img_path.name, 'org_img': org_img, 'org_shape': (h, w)}
        
        outputs = {}

        label, boxes_yolo = get_label(Path(ann_path))
        boxes = self.xywhn2xyxy(boxes_yolo, w, h)
        
        out = self.transform(self.conf_augmentation)(image=img, bbox=np.concatenate((boxes, label), axis=-1))
        assert out['bbox'].shape[-1] == 5  # ltrb + class_label
        outputs.update({'pixel_values': out['image'], 'bbox': out['bbox'][..., :4],
                        'label': torch.as_tensor(out['bbox'][..., 4], dtype=torch.int64)})


        if self._split in ['train', 'training']:
            return outputs

        assert self._split in ['val', 'valid', 'test']
        # outputs.update({'org_img': org_img, 'org_shape': (h, w)})  # TODO: return org_img with batch_size > 1 
        outputs.update({'org_shape': (h, w)})
        return outputs


import torch.nn as nn
import torch.nn.functional as F

from .detection import AnchorGenerator, RPNHead, RegionProposalNetwork, RoIHeads, GeneralizedRCNN, MultiScaleRoIAlign
from .fpn import FPN

IMAGE_SIZE = (512, 512) # TODO: Get from configuration


class FasterRCNN(GeneralizedRCNN):
    def __init__(
        self,
        num_classes,
        intermediate_features_dim,
        # FPN parameters
        fpn_num_outs=4,
        # Anchor parameters
        anchor_sizes=((64,), (128,), (256,), (512,)),
        aspect_ratios=(0.5, 1.0, 2.0),
        # RPN parameters
        rpn_pre_nms_top_n=2000,
        rpn_post_nms_top_n=2000,
        rpn_nms_thresh=0.7,
        rpn_fg_iou_thresh=0.7,
        rpn_bg_iou_thresh=0.3,
        rpn_batch_size_per_image=256,
        rpn_positive_fraction=0.5,
        rpn_score_thresh=0.0,
        # RoI parameters
        roi_output_size=7,
        roi_sampling_ratio=2,
        roi_representation_size=1024,
        # Box parameters
        box_score_thresh=0.05,
        box_nms_thresh=0.5,
        box_detections_per_img=100,
        box_fg_iou_thresh=0.5,
        box_bg_iou_thresh=0.5,
        box_batch_size_per_image=512,
        box_positive_fraction=0.25,
        bbox_reg_weights=None,
        **kwargs,
    ):
        assert fpn_num_outs == len(anchor_sizes)

        neck = FPN(in_channels=intermediate_features_dim, out_channels=intermediate_features_dim[-1], num_outs=fpn_num_outs)

        out_channels = intermediate_features_dim[-1]

        aspect_ratios = (aspect_ratios,) * len(anchor_sizes)
        rpn_anchor_generator = AnchorGenerator(anchor_sizes, aspect_ratios, IMAGE_SIZE)

        rpn_head = RPNHead(out_channels, rpn_anchor_generator.num_anchors_per_location()[0])

        rpn = RegionProposalNetwork(
            rpn_anchor_generator,
            rpn_head,
            rpn_fg_iou_thresh,
            rpn_bg_iou_thresh,
            rpn_batch_size_per_image,
            rpn_positive_fraction,
            rpn_pre_nms_top_n,
            rpn_post_nms_top_n,
            rpn_nms_thresh,
            score_thresh=rpn_score_thresh,
        )

        featmap_names = [str(i) for i in range(neck.num_outs)]
        box_roi_pool = MultiScaleRoIAlign(featmap_names=featmap_names, output_size=roi_output_size, sampling_ratio=roi_sampling_ratio)

        box_head = TwoMLPHead(out_channels * roi_output_size**2, roi_representation_size)
        box_predictor = FastRCNNPredictor(roi_representation_size, num_classes)

        roi_heads = RoIHeads(
            num_classes,
            # Box
            box_roi_pool,
            box_head,
            box_predictor,
            box_fg_iou_thresh,
            box_bg_iou_thresh,
            box_batch_size_per_image,
            box_positive_fraction,
            bbox_reg_weights,
            box_score_thresh,
            box_nms_thresh,
            box_detections_per_img,
        )

        super().__init__(neck, rpn, roi_heads, IMAGE_SIZE)


class TwoMLPHead(nn.Module):
    """
    Standard heads for FPN-based models

    Args:
        in_channels (int): number of input channels
        representation_size (int): size of the intermediate representation
    """

    def __init__(self, in_channels, representation_size):
        super().__init__()

        self.fc6 = nn.Linear(in_channels, representation_size)
        self.fc7 = nn.Linear(representation_size, representation_size)

    def forward(self, x):
        x = x.flatten(start_dim=1)

        x = F.relu(self.fc6(x))
        x = F.relu(self.fc7(x))

        return x


class FastRCNNPredictor(nn.Module):
    """
    Standard classification + bounding box regression layers
    for Fast R-CNN.

    Args:
        in_channels (int): number of input channels
        num_classes (int): number of output classes (including background)
    """

    def __init__(self, in_channels, num_classes):
        super().__init__()
        self.cls_score = nn.Linear(in_channels, num_classes)
        self.bbox_pred = nn.Linear(in_channels, num_classes * 4)

    def forward(self, x):
        x = x.flatten(start_dim=1)
        scores = self.cls_score(x)
        bbox_deltas = self.bbox_pred(x)

        return scores, bbox_deltas


def faster_rcnn(num_classes, intermediate_features_dim, **kwargs):
    configuration = {
        # FPN parameters
        'fpn_num_outs': 4,
        # Anchor parameters
        'anchor_sizes': ((64,), (128,), (256,), (512,)),
        'aspect_ratios': (0.5, 1.0, 2.0),
        # RPN parameters
        'rpn_pre_nms_top_n': 2000,
        'rpn_post_nms_top_n': 2000,
        'rpn_nms_thresh': 0.7,
        'rpn_fg_iou_thresh': 0.7,
        'rpn_bg_iou_thresh': 0.3,
        'rpn_batch_size_per_image': 256,
        'rpn_positive_fraction': 0.5,
        'rpn_score_thresh': 0.0,
        # RoI parameters
        'roi_output_size': 7,
        'roi_sampling_ratio': 2,
        'roi_representation_size': 1024,
        # Box parameters
        'box_score_thresh': 0.05,
        'box_nms_thresh': 0.5,
        'box_detections_per_img': 100,
        'box_fg_iou_thresh': 0.5,
        'box_bg_iou_thresh': 0.5,
        'box_batch_size_per_image': 512,
        'box_positive_fraction': 0.25,
        'bbox_reg_weights': None,
    }
    return FasterRCNN(num_classes=num_classes, intermediate_features_dim=intermediate_features_dim, **configuration)



from .classification import LabelSmoothingCrossEntropy, SoftTargetCrossEntropy
from .common import CrossEntropyLoss
from .detection import RoiHeadLoss, RPNLoss, YOLOXLoss
from .segmentation import BoundaryLoss, PIDNetBoundaryAwareCrossEntropy, PIDNetCrossEntropy

LOSS_DICT = {
    'cross_entropy': CrossEntropyLoss,
    'soft_target_cross_entropy': SoftTargetCrossEntropy,
    'label_smoothing_cross_entropy': LabelSmoothingCrossEntropy,
    'pidnet_cross_entropy': PIDNetCrossEntropy,
    'boundary_loss': BoundaryLoss,
    'pidnet_cross_entropy_with_boundary': PIDNetBoundaryAwareCrossEntropy,
    'roi_head_loss': RoiHeadLoss,
    'rpn_loss': RPNLoss,
    'yolox_loss': YOLOXLoss,
}

PHASE_LIST = ['train', 'valid', 'test']


import json
import os
from itertools import chain
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from omegaconf import DictConfig, ListConfig
from torch.utils.data import random_split

from ..base import BaseDataSampler
from ..utils.constants import IMG_EXTENSIONS
from ..utils.misc import natural_key


def as_tuple(tuple_string: str) -> Tuple:
    tuple_string = tuple_string.replace("(", "")
    tuple_string = tuple_string.replace(")", "")
    tuple_value = tuple(map(int, tuple_string.split(",")))
    return tuple_value


def load_custom_class_map(id_mapping: Union[ListConfig, DictConfig]) -> Tuple[Dict[int, str], Dict[Union[int, Tuple], int]]:
    if isinstance(id_mapping, ListConfig):
        assert isinstance(id_mapping[0], str), f"Unknown type for class name! {type(id_mapping[0])}"
        idx_to_class: Dict[int, str] = dict(enumerate(id_mapping))
        label_value_to_idx = {k: k for k in idx_to_class}
        return idx_to_class, label_value_to_idx

    idx_to_class: Dict[int, str] = {}
    label_value_to_idx: Dict[Union[int, Tuple], int] = {}
    for class_idx, (label_value, class_name) in enumerate(id_mapping.items()):
        assert isinstance(class_name, str), "You need to format id_mapping with key for class_index, value for class_name."
        if isinstance(label_value, (int, tuple)):
            idx_to_class[class_idx] = class_name
            label_value_to_idx[label_value] = class_idx
            continue

        # Check tuple string
        assert isinstance(label_value, str) and label_value.strip().startswith("(") and label_value.strip().endswith(")"), \
            f"Unknown type for color index! Should be one of (int, tuple-style str, tuple)... but {type(label_value)}"
        label_value_tuple: Tuple[int, int, int] = as_tuple(label_value)
        idx_to_class[class_idx] = class_name
        label_value_to_idx[label_value_tuple] = class_idx

    return idx_to_class, label_value_to_idx


class SegmentationDataSampler(BaseDataSampler):
    def __init__(self, conf_data, train_valid_split_ratio):
        super(SegmentationDataSampler, self).__init__(conf_data, train_valid_split_ratio)

    def load_data(self, split='train'):
        data_root = Path(self.conf_data.path.root)
        split_dir = self.conf_data.path[split]
        image_dir: Path = data_root / split_dir.image
        annotation_dir: Path = data_root / split_dir.label
        images: List[str] = []
        labels: List[str] = []
        images_and_targets: List[Dict[str, str]] = []
        if split in ['train', 'valid']:
            for ext in IMG_EXTENSIONS:
                images.extend([str(file) for file in chain(image_dir.glob(f'*{ext}'), image_dir.glob(f'*{ext.upper()}'))])
                # TODO: get paired data from regex pattern matching (conf_data.path.pattern)
                labels.extend([str(file) for file in chain(annotation_dir.glob(f'*{ext}'), annotation_dir.glob(f'*{ext.upper()}'))])

            images = sorted(images, key=lambda k: natural_key(k))
            labels = sorted(labels, key=lambda k: natural_key(k))
            images_and_targets.extend([{'image': str(image), 'label': str(label)} for image, label in zip(images, labels)])

        elif split == 'test':
            for ext in IMG_EXTENSIONS:
                images_and_targets.extend([{'image': str(file), 'label': None}
                                           for file in chain(image_dir.glob(f'*{ext}'), image_dir.glob(f'*{ext.upper()}'))])
            images_and_targets = sorted(images_and_targets, key=lambda k: natural_key(k['image']))
        else:
            raise AssertionError(f"split should be either {['train', 'valid', 'test']}")

        return images_and_targets

    def load_samples(self):
        assert self.conf_data.path.train.image is not None
        assert isinstance(self.conf_data.id_mapping, (ListConfig, DictConfig))

        idx_to_class, label_value_to_idx = load_custom_class_map(id_mapping=self.conf_data.id_mapping)

        exists_valid = self.conf_data.path.valid.image is not None
        exists_test = self.conf_data.path.test.image is not None

        valid_samples = None
        test_samples = None

        train_samples = self.load_data(split='train')
        if exists_valid:
            valid_samples = self.load_data(split='valid')
        if exists_test:
            test_samples = self.load_data(split='test')

        if not exists_valid:
            num_train_splitted = int(len(train_samples) * self.train_valid_split_ratio)
            train_samples, valid_samples = random_split(train_samples, [num_train_splitted, len(train_samples) - num_train_splitted],
                                                        generator=torch.Generator().manual_seed(42))

        return train_samples, valid_samples, test_samples, {'idx_to_class': idx_to_class, 'label_value_to_idx': label_value_to_idx}

    def load_huggingface_samples(self):
        from datasets import load_dataset

        cache_dir = self.conf_data.metadata.custom_cache_dir
        root = self.conf_data.metadata.repo
        subset_name = self.conf_data.metadata.subset
        if cache_dir is not None:
            cache_dir = Path(cache_dir)
            Path(cache_dir).mkdir(exist_ok=True, parents=True)
        total_dataset = load_dataset(root, name=subset_name, cache_dir=cache_dir)

        assert isinstance(self.conf_data.metadata.id_mapping, (ListConfig, DictConfig))

        idx_to_class, label_value_to_idx = load_custom_class_map(id_mapping=self.conf_data.metadata.id_mapping)

        exists_valid = 'validation' in total_dataset
        exists_test = 'test' in total_dataset

        train_samples = total_dataset['train']
        valid_samples = None
        if exists_valid:
            valid_samples = total_dataset['validation']
        test_samples = None
        if exists_test:
            test_samples = total_dataset['test']

        if not exists_valid:
            splitted_datasets = train_samples.train_test_split(test_size=(1 - self.train_valid_split_ratio))
            train_samples = splitted_datasets['train']
            valid_samples = splitted_datasets['test']
        return train_samples, valid_samples, test_samples, {'idx_to_class': idx_to_class, 'label_value_to_idx': label_value_to_idx}



"""
Based on the Torchvision implementation of MobileNetV3.
https://pytorch.org/vision/main/_modules/torchvision/models/mobilenetv3.html
"""
from typing import List

import torch
import torch.nn as nn
from torch import Tensor

from ...op.custom import ConvLayer, InvertedResidual
from ...utils import BackboneOutput

__all__ = ['mobilenetv3_small']

SUPPORTING_TASK = ['classification', 'segmentation']


def list_depth(block_info):
    if isinstance(block_info[0], list):
        return 1 + list_depth(block_info[0])
    else:
        return 1


class MobileNetV3(nn.Module):

    def __init__(
        self,
        task: str,
        block_info, # [in_channels, kernel, expended_channels, out_channels, use_se, activation, stride, dilation]
        **kwargs
    ) -> None:
        super(MobileNetV3, self).__init__()

        self.task = task.lower()
        self.use_intermediate_features = self.task in ['segmentation', 'detection']
        norm_type = 'batch_norm'
        act_type = 'hard_swish'

        # building first layer
        firstconv_output_channels = block_info[0][0][0]
        self.conv_first = ConvLayer(
            in_channels=3,
            out_channels=firstconv_output_channels,
            kernel_size=3,
            stride=2,
            norm_type=norm_type,
            act_type=act_type,
        )

        # building inverted residual blocks
        stages: List[nn.Module] = []

        lastconv_input_channels = block_info[-1][-1][3]
        lastconv_output_channels = 6 * lastconv_input_channels
        for stg_idx, stage_info in enumerate(block_info):
            stage: List[nn.Module] = []

            for block in stage_info:
                in_channels = block[0]
                kernel_size = block[1]
                hidden_channels = block[2]
                out_channels = block[3]
                use_se = block[4]
                act_type_b = block[5].lower()
                stride = block[6]
                dilation = block[7]
                stage.append(
                    InvertedResidual(in_channels=in_channels,
                                     hidden_channels=hidden_channels,
                                     out_channels=out_channels,
                                     kernel_size=kernel_size,
                                     stride=stride,
                                     norm_type=norm_type,
                                     act_type=act_type_b,
                                     use_se=use_se,
                                     dilation=dilation)
                )
            
            # add last conv
            if stg_idx == len(block_info) - 1:
                stage.append(
                    ConvLayer(in_channels=lastconv_input_channels,
                              out_channels=lastconv_output_channels,
                              kernel_size=1,
                              norm_type=norm_type,
                              act_type=act_type,)
                )

            stage = nn.Sequential(*stage)
            stages.append(stage)
        
        self.stages = nn.ModuleList(stages)
        self.avgpool = nn.AdaptiveAvgPool2d(1)

        self._feature_dim = lastconv_output_channels
        self._intermediate_features_dim = [s[-1].out_channels for s in self.stages[:-1]]
        self._intermediate_features_dim += [lastconv_output_channels]

        # init weights
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def forward(self, x: Tensor):
        x = self.conv_first(x)

        all_hidden_states = () if self.use_intermediate_features else None
        for stage in self.stages:
            x = stage(x)
            if self.use_intermediate_features:
                all_hidden_states = all_hidden_states + (x, )
        
        if self.use_intermediate_features:
            return BackboneOutput(intermediate_features=all_hidden_states)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)

        return BackboneOutput(last_feature=x)

    @property
    def feature_dim(self):
        return self._feature_dim

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

    def task_support(self, task):
        return task.lower() in SUPPORTING_TASK


def mobilenetv3_small(task, conf_model_backbone) -> MobileNetV3:
    return MobileNetV3(task, **conf_model_backbone)



from typing import Dict, List, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torchvision.ops import boxes as box_ops

from ...models.heads.detection.experimental.detection import _utils as det_utils


class RoiHeadLoss(nn.Module):
    def __init__(self) -> None:
        super().__init__()
    
    @staticmethod
    def forward(out: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        class_logits, box_regression, labels, regression_targets =\
            out['class_logits'], out['box_regression'], out['labels'], out['regression_targets']

        labels = torch.cat(labels, dim=0)
        regression_targets = torch.cat(regression_targets, dim=0)

        classification_loss = F.cross_entropy(class_logits, labels)

        # get indices that correspond to the regression targets for
        # the corresponding ground truth labels, to be used with
        # advanced indexing
        sampled_pos_inds_subset = torch.where(labels > 0)[0]
        labels_pos = labels[sampled_pos_inds_subset]
        N, num_classes = class_logits.shape
        box_regression = box_regression.reshape(N, box_regression.size(-1) // 4, 4)

        box_loss = F.smooth_l1_loss(
            box_regression[sampled_pos_inds_subset, labels_pos],
            regression_targets[sampled_pos_inds_subset],
            beta=1 / 9,
            reduction="sum",
        )
        box_loss = box_loss / labels.numel()

        losses = {
            "loss_classifier": classification_loss,
            "loss_box_reg": box_loss
        }
        
        # TODO: return as dict
        return sum(losses.values())
        
class RPNLoss(nn.Module):
    def __init__(self,
                 box_fg_iou_thresh=0.5,
                 box_bg_iou_thresh=0.5,
                 box_batch_size_per_image=512,
                 box_positive_fraction=0.25) -> None:
        super().__init__()
        
        self.box_coder = det_utils.BoxCoder(weights=(1.0, 1.0, 1.0, 1.0))
        self.box_similarity = box_ops.box_iou
        self.proposal_matcher = det_utils.Matcher(
            box_fg_iou_thresh,
            box_bg_iou_thresh,
            allow_low_quality_matches=True,
        )
        self.fg_bg_sampler = det_utils.BalancedPositiveNegativeSampler(box_batch_size_per_image, box_positive_fraction)
    
    def _assign_targets_to_anchors(self, anchors: List[Tensor], targets: List[Dict[str, Tensor]]
    ) -> Tuple[List[Tensor], List[Tensor]]:

        labels = []
        matched_gt_boxes = []
        for anchors_per_image, targets_per_image in zip(anchors, targets):
            gt_boxes = targets_per_image["boxes"]

            if gt_boxes.numel() == 0:
                # Background image (negative example)
                device = anchors_per_image.device
                matched_gt_boxes_per_image = torch.zeros(anchors_per_image.shape, dtype=torch.float32, device=device)
                labels_per_image = torch.zeros((anchors_per_image.shape[0],), dtype=torch.float32, device=device)
            else:
                match_quality_matrix = self.box_similarity(gt_boxes, anchors_per_image)
                matched_idxs = self.proposal_matcher(match_quality_matrix)
                # get the targets corresponding GT for each proposal
                # NB: need to clamp the indices because we can have a single
                # GT in the image, and matched_idxs can be -2, which goes
                # out of bounds
                matched_gt_boxes_per_image = gt_boxes[matched_idxs.clamp(min=0)]

                labels_per_image = matched_idxs >= 0
                labels_per_image = labels_per_image.to(dtype=torch.float32)

                # Background (negative examples)
                bg_indices = matched_idxs == self.proposal_matcher.BELOW_LOW_THRESHOLD
                labels_per_image[bg_indices] = 0.0

                # discard indices that are between thresholds
                inds_to_discard = matched_idxs == self.proposal_matcher.BETWEEN_THRESHOLDS
                labels_per_image[inds_to_discard] = -1.0

            labels.append(labels_per_image)
            matched_gt_boxes.append(matched_gt_boxes_per_image)
        return labels, matched_gt_boxes
        
    def _compute_loss(self, objectness: Tensor, pred_bbox_deltas: Tensor, labels: List[Tensor], regression_targets: List[Tensor]
    ) -> Tuple[Tensor, Tensor]:
        """
        Args:
            objectness (Tensor)
            pred_bbox_deltas (Tensor)
            labels (List[Tensor])
            regression_targets (List[Tensor])

        Returns:
            objectness_loss (Tensor)
            box_loss (Tensor)
        """

        sampled_pos_inds, sampled_neg_inds = self.fg_bg_sampler(labels)
        sampled_pos_inds = torch.where(torch.cat(sampled_pos_inds, dim=0))[0]
        sampled_neg_inds = torch.where(torch.cat(sampled_neg_inds, dim=0))[0]

        sampled_inds = torch.cat([sampled_pos_inds, sampled_neg_inds], dim=0)

        objectness = objectness.flatten()

        labels = torch.cat(labels, dim=0)
        regression_targets = torch.cat(regression_targets, dim=0)

        pred_bbox_deltas = pred_bbox_deltas.view(-1, 4)
        box_loss = F.smooth_l1_loss(
            pred_bbox_deltas[sampled_pos_inds],
            regression_targets[sampled_pos_inds],
            beta=1 / 9,
            reduction="sum",
        ) / (sampled_inds.numel())

        objectness_loss = F.binary_cross_entropy_with_logits(objectness[sampled_inds], labels[sampled_inds])

        return objectness_loss, box_loss
    
    def forward(self, out: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        anchors, objectness, pred_bbox_deltas = out['anchors'], out['objectness'], out['pred_bbox_deltas']
        labels, matched_gt_boxes = self._assign_targets_to_anchors(anchors, target)
        regression_targets = self.box_coder.encode(matched_gt_boxes, list(anchors))
        loss_objectness, loss_rpn_box_reg = self._compute_loss(
            objectness, pred_bbox_deltas, labels, regression_targets
        )
        losses = {
            "loss_objectness": loss_objectness,
            "loss_rpn_box_reg": loss_rpn_box_reg,
        }
        # TODO: return as dict
        return sum(losses.values())


from typing import Dict, Type

import torch
from torch.optim.lr_scheduler import _LRScheduler

from .cosine_lr import CosineAnnealingLRWithCustomWarmUp
from .cosine_warm_restart import CosineAnnealingWarmRestartsWithCustomWarmUp
from .poly_lr import PolynomialLRWithWarmUp
from .step_lr import StepLR

SCHEDULER_DICT: Dict[str, Type[_LRScheduler]] = {
    'cosine': CosineAnnealingWarmRestartsWithCustomWarmUp,
    'cosine_no_sgdr': CosineAnnealingLRWithCustomWarmUp,
    'poly': PolynomialLRWithWarmUp,
    'step': StepLR
}


from typing import Dict, List, Optional, Tuple

import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torchvision.ops import boxes as box_ops

from . import _utils as det_utils


class RoIHeads(nn.Module):
    __annotations__ = {
        "box_coder": det_utils.BoxCoder,
        "proposal_matcher": det_utils.Matcher,
        "fg_bg_sampler": det_utils.BalancedPositiveNegativeSampler,
    }

    def __init__(
        self,
        num_classes,
        box_roi_pool,
        box_head,
        box_predictor,
        # Faster R-CNN training
        fg_iou_thresh,
        bg_iou_thresh,
        batch_size_per_image,
        positive_fraction,
        bbox_reg_weights,
        # Faster R-CNN inference
        score_thresh,
        nms_thresh,
        detections_per_img,
        # Mask
        mask_roi_pool=None,
        mask_head=None,
        mask_predictor=None,
        keypoint_roi_pool=None,
        keypoint_head=None,
        keypoint_predictor=None,
    ):
        super().__init__()

        self.class_ids = list(range(num_classes))

        self.box_similarity = box_ops.box_iou
        # assign ground-truth boxes for each proposal
        self.proposal_matcher = det_utils.Matcher(fg_iou_thresh, bg_iou_thresh, allow_low_quality_matches=False)

        self.fg_bg_sampler = det_utils.BalancedPositiveNegativeSampler(batch_size_per_image, positive_fraction)

        if bbox_reg_weights is None:
            bbox_reg_weights = (10.0, 10.0, 5.0, 5.0)
        self.box_coder = det_utils.BoxCoder(bbox_reg_weights)

        self.box_roi_pool = box_roi_pool
        self.box_head = box_head
        self.box_predictor = box_predictor

        self.score_thresh = score_thresh
        self.nms_thresh = nms_thresh
        self.detections_per_img = detections_per_img

        self.mask_roi_pool = mask_roi_pool
        self.mask_head = mask_head
        self.mask_predictor = mask_predictor

        self.keypoint_roi_pool = keypoint_roi_pool
        self.keypoint_head = keypoint_head
        self.keypoint_predictor = keypoint_predictor

    def has_mask(self):
        if self.mask_roi_pool is None:
            return False
        if self.mask_head is None:
            return False
        if self.mask_predictor is None:
            return False
        return True

    def has_keypoint(self):
        if self.keypoint_roi_pool is None:
            return False
        if self.keypoint_head is None:
            return False
        if self.keypoint_predictor is None:
            return False
        return True

    def assign_targets_to_proposals(self, proposals, gt_boxes, gt_labels):
        # type: (List[Tensor], List[Tensor], List[Tensor]) -> Tuple[List[Tensor], List[Tensor]]
        matched_idxs = []
        labels = []
        for proposals_in_image, gt_boxes_in_image, gt_labels_in_image in zip(proposals, gt_boxes, gt_labels):

            if gt_boxes_in_image.numel() == 0:
                # Background image
                device = proposals_in_image.device
                clamped_matched_idxs_in_image = torch.zeros(
                    (proposals_in_image.shape[0],), dtype=torch.int64, device=device
                )
                labels_in_image = torch.zeros((proposals_in_image.shape[0],), dtype=torch.int64, device=device)
            else:
                #  set to self.box_similarity when https://github.com/pytorch/pytorch/issues/27495 lands
                match_quality_matrix = box_ops.box_iou(gt_boxes_in_image, proposals_in_image)
                matched_idxs_in_image = self.proposal_matcher(match_quality_matrix)

                clamped_matched_idxs_in_image = matched_idxs_in_image.clamp(min=0)

                labels_in_image = gt_labels_in_image[clamped_matched_idxs_in_image]
                labels_in_image = labels_in_image.to(dtype=torch.int64)

                # Label background (below the low threshold)
                bg_inds = matched_idxs_in_image == self.proposal_matcher.BELOW_LOW_THRESHOLD
                labels_in_image[bg_inds] = 0

                # Label ignore proposals (between low and high thresholds)
                ignore_inds = matched_idxs_in_image == self.proposal_matcher.BETWEEN_THRESHOLDS
                labels_in_image[ignore_inds] = -1  # -1 is ignored by sampler

            matched_idxs.append(clamped_matched_idxs_in_image)
            labels.append(labels_in_image)
        return matched_idxs, labels

    def subsample(self, labels):
        # type: (List[Tensor]) -> List[Tensor]
        sampled_pos_inds, sampled_neg_inds = self.fg_bg_sampler(labels)
        sampled_inds = []
        for _img_idx, (pos_inds_img, neg_inds_img) in enumerate(zip(sampled_pos_inds, sampled_neg_inds)):
            img_sampled_inds = torch.where(pos_inds_img | neg_inds_img)[0]
            sampled_inds.append(img_sampled_inds)
        return sampled_inds

    def add_gt_proposals(self, proposals, gt_boxes):
        # type: (List[Tensor], List[Tensor]) -> List[Tensor]
        proposals = [torch.cat((proposal, gt_box)) for proposal, gt_box in zip(proposals, gt_boxes)]

        return proposals

    def check_targets(self, targets):
        # type: (Optional[List[Dict[str, Tensor]]]) -> None
        if targets is None:
            raise ValueError("targets should not be None")
        if not all("boxes" in t for t in targets):
            raise ValueError("Every element of targets should have a boxes key")
        if not all("labels" in t for t in targets):
            raise ValueError("Every element of targets should have a labels key")
        if self.has_mask() and not all("masks" in t for t in targets):
            raise ValueError("Every element of targets should have a masks key")

    def select_training_samples(
        self,
        proposals,  # type: List[Tensor]
        targets,  # type: Optional[List[Dict[str, Tensor]]]
    ):
        # type: (...) -> Tuple[List[Tensor], List[Tensor], List[Tensor], List[Tensor]]
        self.check_targets(targets)
        if targets is None:
            raise ValueError("targets should not be None")
        dtype = proposals[0].dtype
        device = proposals[0].device

        gt_boxes = [t["boxes"].to(device, dtype) for t in targets]
        gt_labels = [t["labels"] for t in targets]

        # append ground-truth bboxes to propos
        proposals = self.add_gt_proposals(proposals, gt_boxes)

        # get matching gt indices for each proposal
        matched_idxs, labels = self.assign_targets_to_proposals(proposals, gt_boxes, gt_labels)
        # sample a fixed proportion of positive-negative proposals
        sampled_inds = self.subsample(labels)
        matched_gt_boxes = []
        num_images = len(proposals)
        for img_id in range(num_images):
            img_sampled_inds = sampled_inds[img_id]
            proposals[img_id] = proposals[img_id][img_sampled_inds]
            labels[img_id] = labels[img_id][img_sampled_inds]
            matched_idxs[img_id] = matched_idxs[img_id][img_sampled_inds]

            gt_boxes_in_image = gt_boxes[img_id]
            if gt_boxes_in_image.numel() == 0:
                gt_boxes_in_image = torch.zeros((1, 4), dtype=dtype, device=device)
            matched_gt_boxes.append(gt_boxes_in_image[matched_idxs[img_id]])

        regression_targets = self.box_coder.encode(matched_gt_boxes, proposals)
        return proposals, matched_idxs, labels, regression_targets

    def postprocess_detections(
        self,
        class_logits,  # type: Tensor
        box_regression,  # type: Tensor
        proposals,  # type: List[Tensor]
        image_shapes,  # type: List[Tuple[int, int]]
    ):
        # type: (...) -> Tuple[List[Tensor], List[Tensor], List[Tensor]]
        device = class_logits.device
        num_classes = class_logits.shape[-1]

        boxes_per_image = [boxes_in_image.shape[0] for boxes_in_image in proposals]
        pred_boxes = self.box_coder.decode_single(box_regression, torch.cat(proposals, dim=0))
        pred_boxes = pred_boxes.reshape(box_regression.shape[0], -1, 4)

        pred_scores = F.softmax(class_logits, -1)

        pred_boxes_list = pred_boxes.split(boxes_per_image, 0)
        pred_scores_list = pred_scores.split(boxes_per_image, 0)

        all_boxes = []
        all_scores = []
        all_labels = []

        # Apply Non-maximum suppression
        for boxes, scores in zip(pred_boxes_list, pred_scores_list):
            boxes = box_ops.clip_boxes_to_image(boxes, image_shapes)

            # create labels for each prediction
            labels = torch.arange(num_classes, device=device)
            labels = labels.view(1, -1).expand_as(scores)

            # remove predictions with the background label
            boxes = boxes[:, 1:]
            scores = scores[:, 1:]
            labels = labels[:, 1:]

            # batch everything, by making every class prediction be a separate instance
            boxes = boxes.reshape(-1, 4)
            scores = scores.reshape(-1)
            labels = labels.reshape(-1)

            # remove low scoring boxes
            inds = torch.where(scores > self.score_thresh)[0]
            boxes, scores, labels = boxes[inds], scores[inds], labels[inds]

            # remove empty boxes
            keep = box_ops.remove_small_boxes(boxes, min_size=1e-2)
            boxes, scores, labels = boxes[keep], scores[keep], labels[keep]

            # non-maximum suppression, independently done per class
            keep = box_ops.batched_nms(boxes, scores, labels, self.nms_thresh)
            # keep only topk scoring predictions
            keep = keep[: self.detections_per_img]
            boxes, scores, labels = boxes[keep], scores[keep], labels[keep]

            all_boxes.append(boxes)
            all_scores.append(scores)
            all_labels.append(labels)

        return all_boxes, all_scores, all_labels

    def forward(
        self,
        features,  # type: Dict[str, Tensor]
        proposals,  # type: List[Tensor]
        image_shapes,  # type: Tuple[int, int]
    ):
        # type: (...) -> Tuple[List[Dict[str, Tensor]], Dict[str, Tensor]]
        """
        Args:
            features (List[Tensor])
            proposals (List[Tensor[N, 4]])
            image_shapes (List[Tuple[H, W]])
            targets (List[Dict])
        """
        box_features = self.box_roi_pool(features, proposals, image_shapes)
        box_features = self.box_head(box_features)
        class_logits, box_regression = self.box_predictor(box_features)
        post_boxes, post_scores, post_labels = self.postprocess_detections(class_logits, box_regression, proposals, image_shapes)

        return {
            'class_logits': class_logits,
            'box_regression': box_regression,
            'post_boxes': post_boxes,
            'post_scores': post_scores,
            'post_labels': post_labels
        }


import logging
import os
from typing import Literal

import numpy as np
import torch
from omegaconf import OmegaConf

from .base import BasePipeline

logger = logging.getLogger("netspresso_trainer")

CITYSCAPE_IGNORE_INDEX = 255  # TODO: get from configuration


class SegmentationPipeline(BasePipeline):
    def __init__(self, conf, task, model_name, model, devices, train_dataloader, eval_dataloader, class_map, **kwargs):
        super(SegmentationPipeline, self).__init__(conf, task, model_name, model, devices,
                                                   train_dataloader, eval_dataloader, class_map, **kwargs)
        self.ignore_index = CITYSCAPE_IGNORE_INDEX
        self.num_classes = train_dataloader.dataset.num_classes

    def train_step(self, batch):
        self.model.train()
        images, target = batch['pixel_values'], batch['labels']
        images = images.to(self.devices)
        target = target.long().to(self.devices)

        if 'edges' in batch:
            bd_gt = batch['edges']
            bd_gt = bd_gt.to(self.devices)

        self.optimizer.zero_grad()
        out = self.model(images)
        if 'edges' in batch:
            self.loss_factory.calc(out, target, bd_gt=bd_gt, phase='train')
        else:
            self.loss_factory.calc(out, target, phase='train')

        self.loss_factory.backward()
        self.optimizer.step()

        out = {k: v.detach() for k, v in out.items()}
        self.metric_factory.calc(out['pred'], target, phase='train')

        if self.conf.distributed:
            torch.distributed.barrier()

    def valid_step(self, batch):
        self.model.eval()
        images, target = batch['pixel_values'], batch['labels']
        images = images.to(self.devices)
        target = target.long().to(self.devices)

        if 'edges' in batch:
            bd_gt = batch['edges']
            bd_gt = bd_gt.to(self.devices)

        out = self.model(images)
        if 'edges' in batch:
            self.loss_factory.calc(out, target, bd_gt=bd_gt, phase='valid')
        else:
            self.loss_factory.calc(out, target, phase='valid')

        self.metric_factory.calc(out['pred'], target, phase='valid')

        if self.conf.distributed:
            torch.distributed.barrier()

        output_seg = torch.max(out['pred'], dim=1)[1]  # argmax

        logs = {
            'images': images.detach().cpu().numpy(),
            'target': target.detach().cpu().numpy(),
            'pred': output_seg.detach().cpu().numpy()
        }
        if 'edges' in batch:
            logs.update({
                'bd_gt': bd_gt.detach().cpu().numpy()
            })
        return dict(logs.items())

    def test_step(self, batch):
        self.model.eval()
        images = batch['pixel_values']
        images = images.to(self.devices)

        out = self.model(images.unsqueeze(0))

        output_seg = torch.max(out['pred'], dim=1)[1]  # argmax

        return output_seg

    def get_metric_with_all_outputs(self, outputs, phase: Literal['train', 'valid']):
        pass



from dataclasses import dataclass

from omegaconf import MISSING, MissingMandatoryValue


@dataclass
class ScheduleConfig:
    seed: int = 1
    opt: str = "adamw"
    lr: float = 6e-5
    momentum: float =  0.937
    weight_decay: float = 0.0005
    sched: str = "cosine"
    min_lr: float = 1e-6
    warmup_bias_lr: float = 1e-5
    warmup_epochs: int = 5
    iters_per_phase: int = 30
    sched_power: float = 1.0
    epochs: int = 3
    batch_size: int = 8


@dataclass
class ClassificationScheduleConfig(ScheduleConfig):
    batch_size: int = 32


@dataclass
class SegmentationScheduleConfig(ScheduleConfig):
    pass


@dataclass
class DetectionScheduleConfig(ScheduleConfig):
    pass


import copy
import logging
import os
from abc import ABC, abstractmethod
from dataclasses import asdict
from pathlib import Path
from statistics import mean
from typing import Dict, Literal, final

import torch
import torch.nn as nn
from tqdm import tqdm

from ..loggers import START_EPOCH_ZERO_OR_ONE, build_logger
from ..losses import build_losses
from ..metrics import build_metrics
from ..optimizers import build_optimizer
from ..schedulers import build_scheduler
from ..utils.fx import save_graphmodule
from ..utils.logger import yaml_for_logging
from ..utils.onnx import save_onnx
from ..utils.record import Timer, TrainingSummary
from ..utils.stats import get_params_and_macs

logger = logging.getLogger("netspresso_trainer")

NUM_SAMPLES = 16


class BasePipeline(ABC):
    def __init__(self, conf, task, model_name, model, devices,
                 train_dataloader, eval_dataloader, class_map,
                 is_graphmodule_training=False, profile=False):
        super(BasePipeline, self).__init__()
        self.conf = conf
        self.task = task
        self.model_name = model_name
        self.save_dtype = next(model.parameters()).dtype
        self.model = model.float()
        self.devices = devices
        self.train_dataloader = train_dataloader
        self.eval_dataloader = eval_dataloader
        self.train_step_per_epoch = len(train_dataloader)
        self.training_history: Dict[int, Dict[
            Literal['train_losses', 'valid_losses', 'train_metrics', 'valid_metrics'], Dict[str, float]
        ]] = {}

        self.timer = Timer()

        self.loss_factory = None
        self.metric_factory = None
        self.optimizer = None
        self.start_epoch_at_one = bool(START_EPOCH_ZERO_OR_ONE)
        self.start_epoch = int(self.start_epoch_at_one)

        self.ignore_index = None
        self.num_classes = None

        self.profile = profile  # TODO: provide torch_tb_profiler for training
        self.is_graphmodule_training = is_graphmodule_training
        self.save_optimizer_state = self.conf.logging.save_optimizer_state

        self.single_gpu_or_rank_zero = (not self.conf.distributed) or (self.conf.distributed and torch.distributed.get_rank() == 0)

        if self.single_gpu_or_rank_zero:
            self.train_logger = build_logger(self.conf, self.task, self.model_name,
                                             step_per_epoch=self.train_step_per_epoch, class_map=class_map,
                                             num_sample_images=NUM_SAMPLES)

    @final
    def _is_ready(self):
        assert self.model is not None, "`self.model` is not defined!"
        assert self.optimizer is not None, "`self.optimizer` is not defined!"
        """Append here if you need more assertion checks!"""
        assert self.conf.logging.save_checkpoint_epoch % self.conf.logging.validation_epoch == 0, \
            "`save_checkpoint_epoch` should be the multiplier of `validation_epoch`."
        return True

    def set_train(self):

        assert self.model is not None
        self.optimizer = build_optimizer(self.model,
                                         opt=self.conf.training.opt,
                                         lr=self.conf.training.lr,
                                         wd=self.conf.training.weight_decay,
                                         momentum=self.conf.training.momentum)
        self.scheduler, _ = build_scheduler(self.optimizer, self.conf.training)
        self.loss_factory = build_losses(self.conf.model, ignore_index=self.ignore_index)
        self.metric_factory = build_metrics(self.task, self.conf.model, ignore_index=self.ignore_index, num_classes=self.num_classes)
        resume_optimizer_checkpoint = self.conf.model.resume_optimizer_checkpoint
        if resume_optimizer_checkpoint is not None:
            resume_optimizer_checkpoint = Path(resume_optimizer_checkpoint)
            if not resume_optimizer_checkpoint.exists():
                logger.warning(f"Traning summary checkpoint path {str(resume_optimizer_checkpoint)} is not found!"
                               f"Skip loading the previous history and trainer will be started from the beginning")
                return

            optimizer_dict = torch.load(resume_optimizer_checkpoint, map_location='cpu')
            optimizer_state_dict = optimizer_dict['optimizer']
            start_epoch = optimizer_dict['last_epoch'] + 1  # Start from the next to the end of last training
            start_epoch_at_one = optimizer_dict['start_epoch_at_one']

            self.optimizer.load_state_dict(optimizer_state_dict)
            self.scheduler.step(epoch=start_epoch)

            self.start_epoch_at_one = start_epoch_at_one
            self.start_epoch = start_epoch
            logger.info(f"Resume training from {str(resume_optimizer_checkpoint)}. Start training at epoch: {self.start_epoch}")

    def epoch_with_valid_logging(self, epoch: int):
        validation_freq = self.conf.logging.validation_epoch
        last_epoch = epoch == (self.conf.training.epochs + self.start_epoch_at_one - 1)
        return (epoch % validation_freq == self.start_epoch_at_one % validation_freq) or last_epoch

    def epoch_with_checkpoint_saving(self, epoch: int):
        checkpoint_freq = self.conf.logging.save_checkpoint_epoch
        last_epoch = epoch == (self.conf.training.epochs + self.start_epoch_at_one - 1)
        return (epoch % checkpoint_freq == self.start_epoch_at_one % checkpoint_freq) or last_epoch

    @abstractmethod
    def train_step(self, batch):
        raise NotImplementedError

    @abstractmethod
    def valid_step(self, batch):
        raise NotImplementedError

    @abstractmethod
    def test_step(self, batch):
        raise NotImplementedError

    @abstractmethod
    def get_metric_with_all_outputs(self, outputs, phase: Literal['train', 'valid']):
        raise NotImplementedError

    @property
    def learning_rate(self):
        return mean([param_group['lr'] for param_group in self.optimizer.param_groups])

    @property
    def train_loss(self):
        return self.loss_factory.result('train').get('total').avg

    @property
    def valid_loss(self):
        return self.loss_factory.result('valid').get('total').avg

    @property
    def sample_input(self):
        return torch.randn((1, 3, self.conf.augmentation.img_size, self.conf.augmentation.img_size))

    def train(self):
        logger.debug(f"Training configuration:\n{yaml_for_logging(self.conf)}")
        logger.info("-" * 40)

        self.timer.start_record(name='train_all')
        self._is_ready()

        num_epoch = -1
        try:
            for num_epoch in range(self.start_epoch, self.conf.training.epochs + self.start_epoch_at_one):
                self.timer.start_record(name=f'train_epoch_{num_epoch}')
                self.loss_factory.reset_values()
                self.metric_factory.reset_values()

                self.train_one_epoch()

                with_valid_logging = self.epoch_with_valid_logging(num_epoch)
                with_checkpoint_saving = self.epoch_with_checkpoint_saving(num_epoch)
                # FIXME: multi-gpu sample counting & validation
                valid_samples = self.validate() if with_valid_logging else None

                self.timer.end_record(name=f'train_epoch_{num_epoch}')
                time_for_epoch = self.timer.get(name=f'train_epoch_{num_epoch}', as_pop=False)

                if self.single_gpu_or_rank_zero:
                    self.log_end_epoch(epoch=num_epoch,
                                       time_for_epoch=time_for_epoch,
                                       valid_samples=valid_samples,
                                       valid_logging=with_valid_logging)
                    if with_checkpoint_saving:
                        assert with_valid_logging
                        self.save_checkpoint(epoch=num_epoch)
                        self.save_summary()

                self.scheduler.step()  # call after reporting the current `learning_rate`
                logger.info("-" * 40)

            self.timer.end_record(name='train_all')
            total_train_time = self.timer.get(name='train_all', as_pop=False)
            logger.info(f"Total time: {total_train_time:.2f} s")

            if self.single_gpu_or_rank_zero:
                self.train_logger.log_end_of_traning(final_metrics={'time_for_last_epoch': time_for_epoch})
                self.save_summary(end_training=True)
        except KeyboardInterrupt as e:
            # TODO: add independent procedure for KeyboardInterupt
            logger.error("Keyboard interrupt detected! Try saving the current checkpoint...")
            if self.single_gpu_or_rank_zero:
                self.save_checkpoint(epoch=num_epoch)
                self.save_summary()
            raise e
        except Exception as e:
            logger.error(str(e))
            raise e

    def train_one_epoch(self):
        outputs = []
        for _idx, batch in enumerate(tqdm(self.train_dataloader, leave=False)):
            out = self.train_step(batch)
            outputs.append(out)
        self.get_metric_with_all_outputs(outputs, phase='train')

    @torch.no_grad()
    def validate(self, num_samples=NUM_SAMPLES):
        num_returning_samples = 0
        returning_samples = []
        outputs = []
        for _idx, batch in enumerate(tqdm(self.eval_dataloader, leave=False)):
            out = self.valid_step(batch)
            if out is not None:
                outputs.append(out)
                if num_returning_samples < num_samples:
                    returning_samples.append(out)
                    num_returning_samples += len(out['pred'])
        self.get_metric_with_all_outputs(outputs, phase='valid')
        return returning_samples

    @torch.no_grad()
    def inference(self, test_dataset):
        returning_samples = []
        for _idx, batch in enumerate(tqdm(test_dataset, leave=False)):
            out = self.test_step(batch)
            returning_samples.append(out)
        return returning_samples

    def log_end_epoch(self, epoch, time_for_epoch, valid_samples=None, valid_logging=False):
        train_losses = self.loss_factory.result('train')
        train_metrics = self.metric_factory.result('train')

        valid_losses = self.loss_factory.result('valid') if valid_logging else None
        valid_metrics = self.metric_factory.result('valid') if valid_logging else None

        self.train_logger.update_epoch(epoch)
        self.train_logger.log(
            train_losses=train_losses,
            train_metrics=train_metrics,
            valid_losses=valid_losses,
            valid_metrics=valid_metrics,
            train_images=None,
            valid_images=valid_samples,
            learning_rate=self.learning_rate,
            elapsed_time=time_for_epoch
        )

        summary_record = {'train_losses': train_losses, 'train_metrics': train_metrics}
        if valid_logging:
            summary_record.update({'valid_losses': valid_losses, 'valid_metrics': valid_metrics})
        self.training_history.update({epoch: summary_record})

    def save_checkpoint(self, epoch: int):

        # Check whether the valid loss is minimum at this epoch
        valid_losses = {epoch: record['valid_losses'].get('total') for epoch, record in self.training_history.items()
                        if 'valid_losses' in record}
        best_epoch = min(valid_losses, key=valid_losses.get)
        save_best_model = best_epoch == epoch

        model = self.model.module if hasattr(self.model, 'module') else self.model
        if self.save_dtype == torch.float16:
            model = copy.deepcopy(model).type(self.save_dtype)
        result_dir = self.train_logger.result_dir
        model_path = Path(result_dir) / f"{self.task}_{self.model_name}_epoch_{epoch}.ext"
        best_model_path = Path(result_dir) / f"{self.task}_{self.model_name}_best.ext"
        optimizer_path = Path(result_dir) / f"{self.task}_{self.model_name}_epoch_{epoch}_optimzer.pth"

        if self.save_optimizer_state:
            optimizer = self.optimizer.module if hasattr(self.optimizer, 'module') else self.optimizer
            save_dict = {'optimizer': optimizer.state_dict(), 'start_epoch_at_one': self.start_epoch_at_one, 'last_epoch': epoch}
            torch.save(save_dict, optimizer_path)
            logger.debug(f"Optimizer state saved at {str(optimizer_path)}")

        if self.is_graphmodule_training:
            # Just save graphmodule checkpoint
            torch.save(model, model_path.with_suffix(".pt"))
            logger.debug(f"PyTorch FX model saved at {str(model_path.with_suffix('.pt'))}")
            if save_best_model:
                save_onnx(model, best_model_path.with_suffix(".onnx"), sample_input=self.sample_input.type(self.save_dtype))
                logger.info(f"ONNX model converting and saved at {str(best_model_path.with_suffix('.onnx'))}")
                torch.save(model, best_model_path.with_suffix(".pt"))
                logger.info(f"Best model saved at {str(best_model_path.with_suffix('.pt'))}")
            return
        torch.save(model.state_dict(), model_path.with_suffix(".pth"))
        logger.debug(f"PyTorch model saved at {str(model_path.with_suffix('.pth'))}")
        if save_best_model:
            torch.save(model.state_dict(), best_model_path.with_suffix(".pth"))
            logger.info(f"Best model saved at {str(best_model_path.with_suffix('.pth'))}")

            try:
                save_onnx(model, best_model_path.with_suffix(".onnx"), sample_input=self.sample_input.type(self.save_dtype))
                logger.info(f"ONNX model converting and saved at {str(best_model_path.with_suffix('.onnx'))}")

                save_graphmodule(model, (model_path.parent / f"{best_model_path.stem}_fx").with_suffix(".pt"))
                logger.info(f"PyTorch FX model tracing and saved at {str(best_model_path.with_suffix('.pt'))}")
            except Exception as e:
                logger.error(e)
                pass

    def save_summary(self, end_training=False):
        training_summary = TrainingSummary(
            total_epoch=self.conf.training.epochs,
            start_epoch_at_one=self.start_epoch_at_one,
            train_losses={epoch: record['train_losses'].get('total') for epoch, record in self.training_history.items()},
            valid_losses={epoch: record['valid_losses'].get('total') for epoch, record in self.training_history.items()
                          if 'valid_losses' in record},
            train_metrics={epoch: record['train_metrics'] for epoch, record in self.training_history.items()},
            valid_metrics={epoch: record['valid_metrics'] for epoch, record in self.training_history.items()
                           if 'valid_metrics' in record},
            metrics_list=self.metric_factory.metric_names,
            primary_metric=self.metric_factory.primary_metric,
        )
        if end_training:
            total_train_time = self.timer.get(name='train_all', as_pop=True)
            macs, params = get_params_and_macs(self.model, self.sample_input.float())
            logger.info(f"[Model stats] Params: {(params/1e6):.2f}M | MACs: {(macs/1e9):.2f}G")
            training_summary.total_train_time = total_train_time
            training_summary.macs = macs
            training_summary.params = params

        result_dir = self.train_logger.result_dir
        summary_path = Path(result_dir) / "training_summary.ckpt"
        torch.save(asdict(training_summary), summary_path)
        logger.info(f"Model training summary saved at {str(summary_path)}")

    def profile_one_epoch(self):
        PROFILE_WAIT = 1
        PROFILE_WARMUP = 1
        PROFILE_ACTIVE = 10
        PROFILE_REPEAT = 1
        _ = torch.ones(1).to(self.devices)
        with torch.profiler.profile(
            schedule=torch.profiler.schedule(wait=PROFILE_WAIT,
                                             warmup=PROFILE_WARMUP,
                                             active=PROFILE_ACTIVE,
                                             repeat=PROFILE_REPEAT),
            on_trace_ready=torch.profiler.tensorboard_trace_handler('./log/test'),
            record_shapes=True,
            profile_memory=True,
            with_flops=True,
            with_modules=True
        ) as prof:
            for idx, batch in enumerate(self.train_dataloader):
                if idx >= (PROFILE_WAIT + PROFILE_WARMUP + PROFILE_ACTIVE) * PROFILE_REPEAT:
                    break
                self.train_step(batch)
                prof.step()



import copy
import logging
import os
from pathlib import Path
from typing import Literal

import numpy as np
import torch
import torchvision
from omegaconf import OmegaConf

from ..models import build_model
from ..models.utils import DetectionModelOutput, load_from_checkpoint
from ..utils.fx import save_graphmodule
from ..utils.onnx import save_onnx
from .base import BasePipeline

logger = logging.getLogger("netspresso_trainer")


class TwoStageDetectionPipeline(BasePipeline):
    def __init__(self, conf, task, model_name, model, devices, train_dataloader, eval_dataloader, class_map, **kwargs):
        super(TwoStageDetectionPipeline, self).__init__(conf, task, model_name, model, devices,
                                                train_dataloader, eval_dataloader, class_map, **kwargs)
        self.num_classes = train_dataloader.dataset.num_classes

        # Re-compose torch.fx backbone and nn.Module head
        # To load head weights, config should have head_checkpoint value.
        if kwargs['is_graphmodule_training']:
            model = build_model(conf.model, task, self.num_classes, None, conf.augmentation.img_size)
            model.backbone = self.model
            model.head = load_from_checkpoint(model.head, conf.model.head_checkpoint)
            model = model.to(device=devices)
            self.model = model

    def train_step(self, batch):
        self.model.train()
        images, labels, bboxes = batch['pixel_values'], batch['label'], batch['bbox']
        images = images.to(self.devices)
        targets = [{"boxes": box.to(self.devices), "labels": label.to(self.devices)}
                   for box, label in zip(bboxes, labels)]

        self.optimizer.zero_grad()

        # forward to rpn
        backbone = self.model.backbone
        head = self.model.head

        features = backbone(images)['intermediate_features']
        if head.neck:
            features = head.neck(features)

        features = {str(k): v for k, v in enumerate(features)}
        rpn_features = head.rpn(features, head.image_size)

        # generate proposals for training
        proposals = rpn_features['boxes']
        proposals, matched_idxs, roi_head_labels, regression_targets = head.roi_heads.select_training_samples(proposals, targets)

        # forward to roi head
        roi_features = head.roi_heads(features, proposals, head.image_size)

        # set out
        out = DetectionModelOutput()
        out.update(rpn_features)
        out.update(roi_features)
        out.update({'labels': roi_head_labels, 'regression_targets': regression_targets})

        # Compute loss
        self.loss_factory.calc(out, target=targets, phase='train')

        self.loss_factory.backward()
        self.optimizer.step()

        if self.conf.distributed:
            torch.distributed.barrier()

    def valid_step(self, batch):
        self.model.eval()
        images, labels, bboxes = batch['pixel_values'], batch['label'], batch['bbox']
        bboxes = [bbox.to(self.devices) for bbox in bboxes]
        labels = [label.to(self.devices) for label in labels]
        images = images.to(self.devices)
        targets = [{"boxes": box, "labels": label} for box, label in zip(bboxes, labels)]

        out = self.model(images)

        # Compute loss
        head = self.model.head
        matched_idxs, roi_head_labels = head.roi_heads.assign_targets_to_proposals(out['boxes'], bboxes, labels)
        matched_gt_boxes = [bbox[idx] for idx, bbox in zip(matched_idxs, bboxes)]
        regression_targets = head.roi_heads.box_coder.encode(matched_gt_boxes, out['boxes'])
        out.update({'labels': roi_head_labels, 'regression_targets': regression_targets})
        self.loss_factory.calc(out, target=targets, phase='valid')

        if self.conf.distributed:
            torch.distributed.barrier()

        logs = {
            'images': images.detach().cpu().numpy(),
            'target': [(bbox.detach().cpu().numpy(), label.detach().cpu().numpy())
                       for bbox, label in zip(bboxes, labels)],
            'pred': [(np.concatenate((bbox.detach().cpu().numpy(), confidence.detach().cpu().numpy()[..., np.newaxis]), axis=-1),
                      label.detach().cpu().numpy())
                     for bbox, confidence, label in zip(out['post_boxes'], out['post_scores'], out['post_labels'])],
        }
        return dict(logs.items())

    def test_step(self, batch):
        self.model.eval()
        images = batch['pixel_values']
        images = images.to(self.devices)

        out = self.model(images.unsqueeze(0))

        results = [(bbox.detach().cpu().numpy(), label.detach().cpu().numpy())
                   for bbox, label in zip(out['post_boxes'], out['post_labels'])],

        return results

    def get_metric_with_all_outputs(self, outputs, phase: Literal['train', 'valid']):
        # TODO: Compute metrics for train phase
        if phase == 'train':
            return

        pred = []
        targets = []
        for output_batch in outputs:
            for detection, class_idx in output_batch['target']:
                target_on_image = {}
                target_on_image['boxes'] = detection
                target_on_image['labels'] = class_idx
                targets.append(target_on_image)

            for detection, class_idx in output_batch['pred']:
                pred_on_image = {}
                pred_on_image['post_boxes'] = detection[..., :4]
                pred_on_image['post_scores'] = detection[..., -1]
                pred_on_image['post_labels'] = class_idx
                pred.append(pred_on_image)
        self.metric_factory.calc(pred, target=targets, phase=phase)
        
    def save_checkpoint(self, epoch: int):

        # Check whether the valid loss is minimum at this epoch
        valid_losses = {epoch: record['valid_losses'].get('total') for epoch, record in self.training_history.items()
                        if 'valid_losses' in record}
        best_epoch = min(valid_losses, key=valid_losses.get)
        save_best_model = best_epoch == epoch

        model = self.model.module if hasattr(self.model, 'module') else self.model
        if self.save_dtype == torch.float16:
            model = copy.deepcopy(model).type(self.save_dtype)
        result_dir = self.train_logger.result_dir
        model_path = Path(result_dir) / f"{self.task}_{self.model_name}_epoch_{epoch}.ext"
        best_model_path = Path(result_dir) / f"{self.task}_{self.model_name}_best.ext"
        optimizer_path = Path(result_dir) / f"{self.task}_{self.model_name}_epoch_{epoch}_optimzer.pth"

        if self.save_optimizer_state:
            optimizer = self.optimizer.module if hasattr(self.optimizer, 'module') else self.optimizer
            save_dict = {'optimizer': optimizer.state_dict(), 'start_epoch_at_one': self.start_epoch_at_one, 'last_epoch': epoch}
            torch.save(save_dict, optimizer_path)
            logger.debug(f"Optimizer state saved at {str(optimizer_path)}")

        if self.is_graphmodule_training:
            # Just save graphmodule checkpoint
            torch.save(model, (model_path.parent / f"{model_path.stem}_backbone").with_suffix(".pth"))
            logger.debug(f"PyTorch FX model saved at {(model_path.parent / f'{model_path.stem}_backbone').with_suffix('.pth')}")
            torch.save(model.head.state_dict(), (model_path.parent / f"{model_path.stem}_head").with_suffix(".pth"))
            logger.info(f"Detection head saved at {(model_path.parent / f'{model_path.stem}_head').with_suffix('.pth')}")
            if save_best_model:
                save_onnx(model, best_model_path.with_suffix(".onnx"), sample_input=self.sample_input.type(self.save_dtype))
                logger.info(f"ONNX model converting and saved at {str(best_model_path.with_suffix('.onnx'))}")

                torch.save(model.backbone, (model_path.parent / f"{best_model_path.stem}_backbone").with_suffix(".pt"))
                logger.info(f"Best model saved at {(model_path.parent / f'{best_model_path.stem}_backbone').with_suffix('.pt')}")
                # save head separately
                torch.save(model.head.state_dict(), (model_path.parent / f"{best_model_path.stem}_head").with_suffix(".pth"))
                logger.info(f"Detection head saved at {(model_path.parent / f'{best_model_path.stem}_head').with_suffix('.pth')}")
            return
        torch.save(model.state_dict(), model_path.with_suffix(".pth"))
        logger.debug(f"PyTorch model saved at {str(model_path.with_suffix('.pth'))}")
        if save_best_model:
            torch.save(model.state_dict(), best_model_path.with_suffix(".pth"))
            logger.info(f"Best model saved at {str(best_model_path.with_suffix('.pth'))}")

            try:
                save_onnx(model, best_model_path.with_suffix(".onnx"), sample_input=self.sample_input.type(self.save_dtype))
                logger.info(f"ONNX model converting and saved at {str(best_model_path.with_suffix('.onnx'))}")

                # fx backbone
                save_graphmodule(model.backbone, (model_path.parent / f"{best_model_path.stem}_backbone_fx").with_suffix(".pt"))
                logger.info(f"PyTorch FX model tracing and saved at {(model_path.parent / f'{best_model_path.stem}_backbone_fx').with_suffix('.pt')}")
                # save head separately
                torch.save(model.head.state_dict(), (model_path.parent / f"{best_model_path.stem}_head").with_suffix(".pth"))
                logger.info(f"Detection head saved at {(model_path.parent / f'{best_model_path.stem}_head').with_suffix('.pth')}")
            except Exception as e:
                logger.error(e)
                pass


class OneStageDetectionPipeline(BasePipeline):
    def __init__(self, conf, task, model_name, model, devices, train_dataloader, eval_dataloader, class_map, **kwargs):
        super(OneStageDetectionPipeline, self).__init__(conf, task, model_name, model, devices,
                                                train_dataloader, eval_dataloader, class_map, **kwargs)
        self.num_classes = train_dataloader.dataset.num_classes

    def train_step(self, batch):
        self.model.train()
        images, labels, bboxes = batch['pixel_values'], batch['label'], batch['bbox']
        images = images.to(self.devices)
        targets = [{"boxes": box.to(self.devices), "labels": label.to(self.devices),}
                   for box, label in zip(bboxes, labels)]
        
        targets = {'gt': targets, 
                   'img_size': images.size(-1), 
                   'num_classes': self.num_classes,}

        self.optimizer.zero_grad()

        out = self.model(images)
        self.loss_factory.calc(out, targets, phase='train')

        self.loss_factory.backward()
        self.optimizer.step()

        # TODO: This step will be moved to postprocessor module
        pred = self.decode_outputs(out, dtype=out[0].type(), stage_strides=[images.shape[-1] // o.shape[-1] for o in out])
        pred = self.postprocess(pred, self.num_classes)

        if self.conf.distributed:
            torch.distributed.barrier()

        logs = {
            'target': [(bbox.detach().cpu().numpy(), label.detach().cpu().numpy())
                       for bbox, label in zip(bboxes, labels)],
            'pred': [(torch.cat([p[:, :4], p[:, 5:6]], dim=-1).detach().cpu().numpy(),
                      p[:, 6].to(torch.int).detach().cpu().numpy()) 
                      if p is not None else (np.array([[]]), np.array([]))
                      for p in pred]
        }
        return dict(logs.items())

    def valid_step(self, batch):
        self.model.eval()
        images, labels, bboxes = batch['pixel_values'], batch['label'], batch['bbox']
        images = images.to(self.devices)
        targets = [{"boxes": box.to(self.devices), "labels": label.to(self.devices)}
                   for box, label in zip(bboxes, labels)]
        
        targets = {'gt': targets, 
                   'img_size': images.size(-1), 
                   'num_classes': self.num_classes,}

        self.optimizer.zero_grad()

        out = self.model(images)
        self.loss_factory.calc(out, targets, phase='valid')

        # TODO: This step will be moved to postprocessor module
        pred = self.decode_outputs(out, dtype=out[0].type(), stage_strides=[images.shape[-1] // o.shape[-1] for o in out])
        pred = self.postprocess(pred, self.num_classes)

        if self.conf.distributed:
            torch.distributed.barrier()

        logs = {
            'images': images.detach().cpu().numpy(),
            'target': [(bbox.detach().cpu().numpy(), label.detach().cpu().numpy())
                       for bbox, label in zip(bboxes, labels)],
            'pred': [(torch.cat([p[:, :4], p[:, 5:6]], dim=-1).detach().cpu().numpy(),
                      p[:, 6].to(torch.int).detach().cpu().numpy()) 
                      if p is not None else (np.array([[]]), np.array([]))
                      for p in pred]
        }
        return dict(logs.items())

    def test_step(self, batch):
        self.model.eval()
        images = batch['pixel_values']
        images = images.to(self.devices)

        out = self.model(images.unsqueeze(0))

        # TODO: This step will be moved to postprocessor module
        pred = self.decode_outputs(out, dtype=out[0].type(), stage_strides=[images.shape[-1] // o.shape[-1] for o in out])
        pred = self.postprocess(pred, self.num_classes)

        results = [(p[:, :4].detach().cpu().numpy(), p[:, 6].to(torch.int).detach().cpu().numpy())
                   if p is not None else (np.array([[]]), np.array([]))
                   for p in pred]

        return results

    def get_metric_with_all_outputs(self, outputs, phase: Literal['train', 'valid']):
        pred = []
        targets = []
        for output_batch in outputs:
            for detection, class_idx in output_batch['target']:
                target_on_image = {}
                target_on_image['boxes'] = detection
                target_on_image['labels'] = class_idx
                targets.append(target_on_image)

            for detection, class_idx in output_batch['pred']:
                pred_on_image = {}
                pred_on_image['post_boxes'] = detection[..., :4]
                pred_on_image['post_scores'] = detection[..., -1]
                pred_on_image['post_labels'] = class_idx
                pred.append(pred_on_image)
        self.metric_factory.calc(pred, target=targets, phase=phase)

    # TODO: Temporary defined in pipeline, it will be moved to postprocessor module.
    def decode_outputs(self, outputs, dtype, stage_strides):
        hw = [x.shape[-2:] for x in outputs]
        # [batch, n_anchors_all, num_classes + 5]
        outputs = torch.cat([x.flatten(start_dim=2) for x in outputs], dim=2).permute(0, 2, 1)
        outputs[..., 4:] = outputs[..., 4:].sigmoid()

        grids = []
        strides = []
        for (hsize, wsize), stride in zip(hw, stage_strides):
            yv, xv = torch.meshgrid(torch.arange(hsize), torch.arange(wsize), indexing='ij')
            grid = torch.stack((xv, yv), 2).view(1, -1, 2)
            grids.append(grid)
            shape = grid.shape[:2]
            strides.append(torch.full((*shape, 1), stride))

        grids = torch.cat(grids, dim=1).type(dtype)
        strides = torch.cat(strides, dim=1).type(dtype)

        outputs = torch.cat([
            (outputs[..., 0:2] + grids) * strides,
            torch.exp(outputs[..., 2:4]) * strides,
            outputs[..., 4:]
        ], dim=-1)
        return outputs

    # TODO: Temporary defined in pipeline, it will be moved to postprocessor module.
    def postprocess(self, prediction, num_classes, conf_thre=0.7, nms_thre=0.45, class_agnostic=False):
        box_corner = prediction.new(prediction.shape)
        box_corner[:, :, 0] = prediction[:, :, 0] - prediction[:, :, 2] / 2
        box_corner[:, :, 1] = prediction[:, :, 1] - prediction[:, :, 3] / 2
        box_corner[:, :, 2] = prediction[:, :, 0] + prediction[:, :, 2] / 2
        box_corner[:, :, 3] = prediction[:, :, 1] + prediction[:, :, 3] / 2
        prediction[:, :, :4] = box_corner[:, :, :4]

        output = [torch.zeros(0, 7).to(prediction.device) for i in range(len(prediction))]
        for i, image_pred in enumerate(prediction):

            # If none are remaining => process next image
            if not image_pred.size(0):
                continue
            # Get score and class with highest confidence
            class_conf, class_pred = torch.max(image_pred[:, 5: 5 + num_classes], 1, keepdim=True)

            conf_mask = (image_pred[:, 4] * class_conf.squeeze() >= conf_thre).squeeze()
            # Detections ordered as (x1, y1, x2, y2, obj_conf, class_conf, class_pred)
            detections = torch.cat((image_pred[:, :5], class_conf, class_pred.float()), 1)
            detections = detections[conf_mask]
            if not detections.size(0):
                continue

            if class_agnostic:
                nms_out_index = torchvision.ops.nms(
                    detections[:, :4],
                    detections[:, 4] * detections[:, 5],
                    nms_thre,
                )
            else:
                nms_out_index = torchvision.ops.batched_nms(
                    detections[:, :4],
                    detections[:, 4] * detections[:, 5],
                    detections[:, 6],
                    nms_thre,
                )

            detections = detections[nms_out_index]
            output[i] = torch.cat((output[i], detections))

        return output



from typing import Union

import torch
import torch.nn as nn
from torch import Tensor
from torch.fx.proxy import Proxy

from ....utils import ModelOutput


class FC(nn.Module):
    def __init__(self, feature_dim: int, num_classes: int) -> None:
        super(FC, self).__init__()
        self.classifier = nn.Linear(feature_dim, num_classes)
        
    def forward(self, x: Union[Tensor, Proxy]) -> ModelOutput:
        x = self.classifier(x)
        return ModelOutput(pred=x)
    
def fc(feature_dim, num_classes, **kwargs):
    return FC(feature_dim=feature_dim, num_classes=num_classes)


from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

CSV_FILENAME = "results.csv"
class BaseCSVLogger(ABC):
    def __init__(self, model, result_dir):
        super(BaseCSVLogger, self).__init__()
        '''
        TODO: if the column name can be changed from the original NP-Searcher,
        a single CSVLogger can be used for all tasks.
        '''
        self.model = model
        self.csv_path = Path(result_dir) / CSV_FILENAME
        self.header: List = []
        
        self._temp_row_dict = {}
        
        if self.csv_path.exists():
            self.csv_path.unlink()
            
        self._epoch = None
        
    @property
    @abstractmethod
    def key_map(self) -> Dict[str, str]:
        raise NotImplementedError
    
    def init_epoch(self):
        self._epoch = 0
        
    @property
    def epoch(self):
        return self._epoch
    
    @epoch.setter
    def epoch(self, value: int) -> None:
        self._epoch = int(value)
        
    def update_header(self, header: List):
        assert len(header) != 0
        self.header = header
        
        with open(self.csv_path, 'a') as f:
            f.write(",".join(self.header))
            f.write("\n")

    def _clear_temp(self):
        self._temp_row_dict = {}
    
    def _update_with_list(self, data: List):
        if data is not None and len(data) != 0:
            with open(self.csv_path, 'a') as f:
                f.write(",".join([f"{x:.09f}" for x in data]))
                f.write("\n")
        self._clear_temp()
        return
    
    def _update_specific(self, data: Dict):
        for _key, _value in data.items():
            if _key not in self.header:
                raise AssertionError(f"The given key ({_key}) is not in {self.header}!")
            if _key not in self._temp_row_dict:
                self._temp_row_dict[_key] = _value
        
        if set(self.header) == set(self._temp_row_dict.keys()):
            self._update_with_list([self._temp_row_dict[_col] for _col in self.header])
        return
            
    def update(self, data=None, **kwargs):
        if isinstance(data, List):
            return self._update_with_list(data)
        if isinstance(data, Dict):
            return self._update_specific(data)
        # if isinstance(data, type(None)):
        #     return self._update_specific(kwargs)
        
        raise AssertionError(f"Type of data should be either List or Dict! Current: {type(data)}")
    
    def _convert_as_csv_record(self, scalar_dict: Dict, prefix: Literal['train', 'valid'] = 'train'):
        converted_dict = {}
        for k, v in scalar_dict.items():
            if f"{prefix}/{k}" not in self.key_map:
                continue
            record_key = self.key_map[f"{prefix}/{k}"]
            assert record_key in self.header, f"{record_key} not in {self.header}"
            
            converted_dict.update({record_key: v})
        return converted_dict
    
    def __call__(self, train_losses, train_metrics, valid_losses=None, valid_metrics=None):
        assert len(self.header) != 0
        assert len(self.key_map) != 0
        
        csv_record_dict = {'epoch': self._epoch}
        converted_train_losses = self._convert_as_csv_record(train_losses, prefix='train')
        converted_train_metrics = self._convert_as_csv_record(train_metrics, prefix='train')
        csv_record_dict.update(converted_train_losses)
        csv_record_dict.update(converted_train_metrics)
        
        if valid_losses is not None:
            converted_valid_losses = self._convert_as_csv_record(valid_losses, prefix='valid')
            csv_record_dict.update(converted_valid_losses)
        if valid_metrics is not None:
            converted_valid_metrics = self._convert_as_csv_record(valid_metrics, prefix='valid')
            csv_record_dict.update(converted_valid_metrics)
        
        self.update(csv_record_dict)



from pathlib import Path
from typing import Union

import torch
import torch.nn as nn
from torch import Tensor

from .environment import get_device

__all__ = ['save_onnx']


def _save_onnx(model: nn.Module, f: Union[str, Path], sample_input: Tensor,
               opset_version=13, input_names='images', output_names='output'):
    torch.onnx.export(model,  # model being run
                      sample_input,  # model input (or a tuple for multiple inputs)
                      f,  # where to save the model (can be a file or file-like object)
                      export_params=True,  # store the trained parameter weights inside the model file
                      opset_version=opset_version,  # the ONNX version to export the model to
                      do_constant_folding=True,  # whether to execute constant folding for optimization
                      input_names=[input_names],  # the model's input names
                      output_names=[output_names],  # the model's output names
                      dynamic_axes={input_names: {0: 'batch_size'},  # variable length axes
                                    output_names: {0: 'batch_size'}})


def save_onnx(model: nn.Module, f: Union[str, Path], sample_input: Tensor):
    sample_input = sample_input.to(get_device(model))
    return _save_onnx(model, f, sample_input, opset_version=13, input_names='images', output_names='output')



from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from omegaconf import MISSING, MissingMandatoryValue

__all__ = [
    "ModelConfig",
    "ClassificationEfficientFormerModelConfig",
    "SegmentationEfficientFormerModelConfig",
    "DetectionEfficientFormerModelConfig",
    "ClassificationMobileNetV3ModelConfig",
    "SegmentationMobileNetV3ModelConfig",
    "ClassificationMobileViTModelConfig",
    "PIDNetModelConfig",
    "ClassificationResNetModelConfig",
    "SegmentationResNetModelConfig",
    "ClassificationSegFormerModelConfig",
    "SegmentationSegFormerModelConfig",
    "ClassificationViTModelConfig",
]

@dataclass
class ArchitectureConfig:
    full: Optional[Dict[str, Any]] = None
    backbone: Optional[Dict[str, Any]] = None
    head: Optional[Dict[str, Any]] = None
    
    def __post_init__(self):
        assert bool(self.full) != bool(self.backbone), "Only one of full or backbone should be given."
    
@dataclass
class ModelConfig:
    task: str = MISSING
    checkpoint: Optional[Union[Path, str]] = None
    fx_model_checkpoint: Optional[Union[Path, str]] = None
    resume_optimizer_checkpoint: Optional[Union[Path, str]] = None
    freeze_backbone: bool = False
    architecture: ArchitectureConfig = field(default_factory=lambda: ArchitectureConfig())
    losses: Optional[List[Dict[str, Any]]] = None


@dataclass
class EfficientFormerArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "efficientformer",
        "num_blocks": [3, 2, 6, 4],
        "hidden_sizes": [48, 96, 224, 448],
        "num_attention_heads": 8,
        "attention_hidden_size": 256,  # attention_hidden_size_splitted * num_attention_heads
        "attention_dropout_prob": 0.,
        "attention_ratio": 4,
        "attention_bias_resolution": 16,
        "pool_size": 3,
        "intermediate_ratio": 4,
        "hidden_dropout_prob": 0.,
        "hidden_activation_type": 'gelu',
        "layer_norm_eps": 1e-5,
        "drop_path_rate": 0.,
        "use_layer_scale": True,
        "layer_scale_init_value": 1e-5,
        "downsamples": [True, True, True, True],
        "down_patch_size": 3,
        "down_stride": 2,
        "down_pad": 1,
        "vit_num": 1,
    })


@dataclass
class MobileNetV3ArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "mobilenetv3_small",
        
        # [in_channels, kernel, expended_channels, out_channels, use_se, activation, stride, dilation]
        "block_info": [
            [
                [16, 3, 16, 16, True, "relu", 2, 1]
            ],
            [
                [16, 3, 72, 24, False, "relu", 2, 1],
                [24, 3, 88, 24, False, "relu", 1, 1]
            ],
            [
                [24, 5, 96, 40, True, "hard_swish", 2, 1],
                [40, 5, 240, 40, True, "hard_swish", 1, 1],
                [40, 5, 240, 40, True, "hard_swish", 1, 1],
                [40, 5, 120, 48, True, "hard_swish", 1, 1],
                [48, 5, 144, 48, True, "hard_swish", 1, 1]
            ],
            [
                [48, 5, 288, 96, True, "hard_swish", 2, 1],
                [96, 5, 576, 96, True, "hard_swish", 1, 1],
                [96, 5, 576, 96, True, "hard_swish", 1, 1]
            ]
        ]
    })


@dataclass
class MobileViTArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "mobilevit",
        "out_channels": [32, 64, 96, 128, 160],
        "block_type": ['mv2', 'mv2', 'mobilevit', 'mobilevit', 'mobilevit'],
        "num_blocks": [1, 3, None, None, None],
        "stride": [1, 2, 2, 2, 2],
        "hidden_size": [None, None, 144, 192, 240],
        "intermediate_size": [None, None, 288, 384, 480],
        "num_transformer_blocks": [None, None, 2, 4, 3],
        "dilate": [None, None, False, False, False],
        "expand_ratio": [4, 4, 4, 4, 4],  # [mv2_exp_mult] * 4
        "patch_embedding_out_channels": 16,
        "local_kernel_size": 3,
        "patch_size": 2,
        "num_attention_heads": 4,  # num_heads
        "attention_dropout_prob": 0.1,
        "hidden_dropout_prob": 0.0,
        "exp_factor": 4,
        "layer_norm_eps": 1e-5,
        "use_fusion_layer": True,
    })


@dataclass
class PIDNetArchitectureConfig(ArchitectureConfig):
    full: Dict[str, Any] = field(default_factory=lambda: {
        "name": "pidnet",
        "m": 2,
        "n": 3,
        "planes": 32,
        "ppm_planes": 96,
        "head_planes": 128,
    })


@dataclass
class ResNetArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "resnet50",
        "block": "bottleneck",
        "layers": [3, 4, 6, 3],
    })


@dataclass
class SegFormerArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "segformer",
        "num_modules": 4,
        "num_blocks": [2, 2, 2, 2],
        "sr_ratios": [8, 4, 2, 1],
        "hidden_sizes": [32, 64, 160, 256],
        "embedding_patch_sizes": [7, 3, 3, 3],
        "embedding_strides": [4, 2, 2, 2],
        "num_attention_heads": [1, 2, 5, 8],
        "intermediate_ratio": 4,
        "hidden_activation_type": "gelu",
        "hidden_dropout_prob": 0.0,
        "attention_dropout_prob": 0.0,
        "layer_norm_eps": 1e-5,
    })


@dataclass
class ViTArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "vit",
        "patch_size": 16,
        "hidden_size": 192,
        "num_blocks": 12,
        "num_attention_heads": 3,
        "attention_dropout_prob": 0.0,
        "intermediate_size": 192 * 4,
        "hidden_dropout_prob": 0.1,
    })


@dataclass
class ClassificationEfficientFormerModelConfig(ModelConfig):
    task: str = "classification"
    checkpoint: Optional[Union[Path, str]] = "./weights/efficientformer/efficientformer_l1_1000d.pth"
    architecture: ArchitectureConfig = field(default_factory=lambda: EfficientFormerArchitectureConfig(
        head={"name": "fc"}
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "label_smoothing_cross_entropy", "smoothing": 0.1, "weight": None}
    ])


@dataclass
class SegmentationEfficientFormerModelConfig(ModelConfig):
    task: str = "segmentation"
    checkpoint: Optional[Union[Path, str]] = "./weights/efficientformer/efficientformer_l1_1000d.pth"
    architecture: ArchitectureConfig = field(default_factory=lambda: EfficientFormerArchitectureConfig(
        head={"name": "all_mlp_decoder"}
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "ignore_index": 255, "weight": None}
    ])


@dataclass
class DetectionEfficientFormerModelConfig(ModelConfig):
    task: str = "detection"
    checkpoint: Optional[Union[Path, str]] = "./weights/efficientformer/efficientformer_l1_1000d.pth"
    architecture: ArchitectureConfig = field(default_factory=lambda: EfficientFormerArchitectureConfig(
        head={"name": "faster_rcnn"}
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "roi_head_loss", "weight": None},
        {"criterion": "rpn_loss", "weight": None},
    ])


@dataclass
class ClassificationMobileNetV3ModelConfig(ModelConfig):
    task: str = "classification"
    checkpoint: Optional[Union[Path, str]] = "./weights/mobilenetv3/mobilenet_v3_small.pth"
    architecture: ArchitectureConfig = field(default_factory=lambda: MobileNetV3ArchitectureConfig(
        head={"name": "fc"}
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "label_smoothing_cross_entropy", "smoothing": 0.1, "weight": None}
    ])


@dataclass
class SegmentationMobileNetV3ModelConfig(ModelConfig):
    task: str = "segmentation"
    checkpoint: Optional[Union[Path, str]] = "./weights/mobilenetv3/mobilenet_v3_small.pth"
    architecture: ArchitectureConfig = field(default_factory=lambda: MobileNetV3ArchitectureConfig(
        head={"name": "all_mlp_decoder"}
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "ignore_index": 255, "weight": None}
    ])


@dataclass
class ClassificationMobileViTModelConfig(ModelConfig):
    task: str = "classification"
    checkpoint: Optional[Union[Path, str]] = "./weights/mobilevit/mobilevit_s.pth"
    architecture: ArchitectureConfig = field(default_factory=lambda: MobileViTArchitectureConfig(
        head={"name": "fc"}
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "label_smoothing_cross_entropy", "smoothing": 0.1, "weight": None}
    ])


@dataclass
class PIDNetModelConfig(ModelConfig):
    task: str = "classification"
    checkpoint: Optional[Union[Path, str]] = "./weights/pidnet/pidnet_s.pth"
    architecture: ArchitectureConfig = field(default_factory=lambda: PIDNetArchitectureConfig())
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "pidnet_cross_entropy", "ignore_index": 255, "weight": None},
        {"criterion": "boundary_loss", "weight": 20.0},
        {"criterion": "pidnet_cross_entropy_with_boundary", "ignore_index": 255, "weight": None},
    ])


@dataclass
class ClassificationResNetModelConfig(ModelConfig):
    task: str = "classification"
    checkpoint: Optional[Union[Path, str]] = "./weights/resnet/resnet50.pth"
    architecture: ArchitectureConfig = field(default_factory=lambda: ResNetArchitectureConfig(
        head={"name": "fc"}
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "label_smoothing_cross_entropy", "smoothing": 0.1, "weight": None}
    ])


@dataclass
class SegmentationResNetModelConfig(ModelConfig):
    task: str = "segmentation"
    checkpoint: Optional[Union[Path, str]] = "./weights/resnet/resnet50.pth"
    architecture: ArchitectureConfig = field(default_factory=lambda: ResNetArchitectureConfig(
        head={"name": "all_mlp_decoder"}
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "ignore_index": 255, "weight": None}
    ])


@dataclass
class ClassificationSegFormerModelConfig(ModelConfig):
    task: str = "classification"
    checkpoint: Optional[Union[Path, str]] = "./weights/segformer/segformer.pth"
    architecture: ArchitectureConfig = field(default_factory=lambda: SegFormerArchitectureConfig(
        head={"name": "fc"}
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "label_smoothing_cross_entropy", "smoothing": 0.1, "weight": None}
    ])


@dataclass
class SegmentationSegFormerModelConfig(ModelConfig):
    task: str = "segmentation"
    checkpoint: Optional[Union[Path, str]] = "./weights/segformer/segformer.pth"
    architecture: ArchitectureConfig = field(default_factory=lambda: SegFormerArchitectureConfig(
        head={"name": "all_mlp_decoder"}
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "ignore_index": 255, "weight": None}
    ])


@dataclass
class ClassificationViTModelConfig(ModelConfig):
    task: str = "classification"
    checkpoint: Optional[Union[Path, str]] = "./weights/vit/vit-tiny.pth"
    architecture: ArchitectureConfig = field(default_factory=lambda: ViTArchitectureConfig(
        head={"name": "fc"}
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "label_smoothing_cross_entropy", "smoothing": 0.1, "weight": None}
    ])




from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F


class SoftTargetCrossEntropy(nn.Module): # cutmix/mixup augmentation
    def __init__(self):
        super(SoftTargetCrossEntropy, self).__init__()

    def forward(self, out: Dict, target: torch.Tensor) -> torch.Tensor:
        pred = out['pred']
        loss = torch.sum(-target * F.log_softmax(pred, dim=-1), dim=-1)
        return loss.mean()


import argparse
import os
import subprocess
from pathlib import Path
from typing import Union

from omegaconf import DictConfig, OmegaConf

from netspresso_trainer.trainer_common import train_common

LOG_LEVEL = os.getenv('LOG_LEVEL', 'INFO')


def run_distributed_training_script(gpu_ids, data, augmentation, model, training, logging, environment, log_level):
    
    command = [
        "--data", data,
        "--augmentation", augmentation, 
        "--model", model,
        "--training", training,
        "--logging", logging,
        "--environment", environment,
        "--log_level", log_level,
    ]
    
    # Distributed training script
    command = [
        'python', '-m', 'torch.distributed.launch',
        f'--nproc_per_node={len(gpu_ids)}',  # GPU #
        os.path.abspath(__file__), *command
    ]

    # Run subprocess
    process = subprocess.Popen(command)

    try:
        process.wait()
    except KeyboardInterrupt:
        print("Interrupted. Terminating the training process...")
        process.terminate()
        process.wait()


def parse_gpu_ids(gpu_arg: str):
    """Parse comma-separated GPU IDs and return as a list of integers."""
    try:
        gpu_ids = [int(id) for id in gpu_arg.split(',')]
        
        if len(gpu_ids) == 1:  # Single GPU
            return gpu_ids[0]
        
        gpu_ids = sorted(gpu_ids)
        return gpu_ids
    except ValueError as e:
        raise argparse.ArgumentTypeError('Invalid GPU IDs. Please provide comma-separated integers.') from e


def parse_args_netspresso(with_gpus=False):

    parser = argparse.ArgumentParser(description="Parser for NetsPresso configuration")

    # -------- User arguments ----------------------------------------
    
    if with_gpus:
        parser.add_argument(
            '--gpus', type=parse_gpu_ids, default=0,
            dest='gpus',
            help='GPU device indices (comma-separated)')

    parser.add_argument(
        '--data', type=str, required=True,
        dest='data',
        help="Config for dataset information")

    parser.add_argument(
        '--augmentation', type=str, required=True,
        dest='augmentation',
        help="Config for data augmentation")

    parser.add_argument(
        '--model', type=str, required=True,
        dest='model',
        help="Config for the model architecture")

    parser.add_argument(
        '--training', type=str, required=True,
        dest='training',
        help="Config for training options")

    parser.add_argument(
        '--logging', type=str, default='config/logging.yaml',
        dest='logging',
        help="Config for logging options")

    parser.add_argument(
        '--environment', type=str, default='config/environment.yaml',
        dest='environment',
        help="Config for training environment (# workers, etc.)")

    parser.add_argument(
        '--log-level', choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'], default=LOG_LEVEL,
        dest='log_level',
        help="Config for training environment (# workers, etc.)")

    args_parsed, _ = parser.parse_known_args()

    return args_parsed


def set_arguments(data: Union[Path, str], augmentation: Union[Path, str],
                  model: Union[Path, str], training: Union[Path, str],
                  logging: Union[Path, str], environment: Union[Path, str]) -> DictConfig:
    
    conf_data = OmegaConf.load(data)
    conf_augmentation = OmegaConf.load(augmentation)
    conf_model = OmegaConf.load(model)
    conf_training = OmegaConf.load(training)
    conf_logging = OmegaConf.load(logging)
    conf_environment = OmegaConf.load(environment)

    conf = OmegaConf.create()
    conf.merge_with(conf_data)
    conf.merge_with(conf_augmentation)
    conf.merge_with(conf_model)
    conf.merge_with(conf_training)
    conf.merge_with(conf_logging)
    conf.merge_with(conf_environment)
    
    return conf


def train_with_yaml_impl(gpus: Union[list, int], data: Union[Path, str], augmentation: Union[Path, str],
                         model: Union[Path, str], training: Union[Path, str],
                         logging: Union[Path, str], environment: Union[Path, str], log_level: str = LOG_LEVEL):
    
    assert isinstance(gpus, (list, int))
    gpu_ids_str = ','.join(map(str, gpus)) if isinstance(gpus, list) else str(gpus)
    os.environ['CUDA_VISIBLE_DEVICES'] = gpu_ids_str
    
    if isinstance(gpus, int):
        conf = set_arguments(data, augmentation, model, training, logging, environment)
        train_common(conf, log_level=log_level)
    else:
        run_distributed_training_script(gpus, data, augmentation, model, training, logging, environment, log_level)


def train_cli() -> None:
    args_parsed = parse_args_netspresso(with_gpus=True)
    
    train_with_yaml_impl(
        gpus=args_parsed.gpus,
        data=args_parsed.data,
        augmentation=args_parsed.augmentation,
        model=args_parsed.model,
        training=args_parsed.training,
        logging=args_parsed.logging,
        environment=args_parsed.environment,
        log_level=args_parsed.log_level
    )


def train_cli_without_additional_gpu_check() -> None:
    args_parsed = parse_args_netspresso(with_gpus=False)
    
    conf = set_arguments(
        data=args_parsed.data,
        augmentation=args_parsed.augmentation,
        model=args_parsed.model,
        training=args_parsed.training,
        logging=args_parsed.logging,
        environment=args_parsed.environment
    )

    train_common(conf, log_level=args_parsed.log_level)


if __name__ == "__main__":
    
    # Execute by `run_distributed_training_script`
    train_cli_without_additional_gpu_check()


""" Inspired from https://github.com/huggingface/pytorch-image-models/blob/main/timm/optim/optim_factory.py
"""
from typing import Callable, Literal, Optional, Tuple

import torch.nn as nn

from .registry import OPTIMIZER_DICT


def build_optimizer(
    model_or_params,
    opt: str = 'adamw',
    lr: Optional[float] = None,
    wd: float = 0.,
    momentum: float = 0.9,
):

    parameters = model_or_params.parameters() if isinstance(model_or_params, nn.Module) else model_or_params

    opt_name: Literal['sgd', 'nesterov', 'momentum',
                      'adam', 'adamw', 'adamax',
                      'adadelta', 'adagrad', 'rmsprop'] = opt.lower()
    assert opt_name in OPTIMIZER_DICT
    
    conf_optim = {'weight_decay': wd, 'lr': lr}

    if opt_name in ['sgd', 'nesterov', 'momentum', 'rmsprop']:
        conf_optim.update({'momentum': momentum})
    if opt_name in ['rmsprop']:
        conf_optim.update({'alpha': 0.9})
    if opt_name in ['sgd', 'nesterov']:
        conf_optim.update({'nesterov': True})
    if opt_name in ['momentum']:
        conf_optim.update({'nesterov': False})
        
    optimizer = OPTIMIZER_DICT[opt_name](parameters, **conf_optim)

    return optimizer
